from __future__ import annotations

import copy
import dataclasses


@dataclasses.dataclass
class MenuItem:
    """
    MenuItem data model.

    ## Attributes

    `name`: The name of the menu item.

    `description`: The description of the menu item.

    `price`: The price of the menu item.

    `category`: The category of the menu item.
    """

    name: str
    description: str
    price: float
    category: str

    @staticmethod
    def new_empty() -> MenuItem:
        """
        Creates a new empty MenuItem object.
        """
        return MenuItem(
            name="",
            description="",
            price=0.0,
            category="",
        )

    def copy(self) -> MenuItem:
        """
        Creates a copy of the MenuItem object.
        """
        return copy.copy(self)



import typing as t

from ..deprecations import deprecated
from ..utils import URL
from .component import Component
from .webview import Webview

__all__ = [
    "Website",
]


@t.final
@deprecated(since="0.10.9", replacement=Webview)
class Website(Component):
    """
    Displays a website.

    `Website` takes a URL as input and displays that website in your app. Since
    the website component cannot know how large its content is going to be,
    you'll probably want to assign a width and height to it.


    ## Attributes
    `url`: The URL of the website you want to display.


    ## Examples

    Here's a simple example that will display an example website:

    ```python
    rio.Website(
        url=rio.URL("https://www.example.com"),
    )
    ```
    """

    url: URL

    def build(self):
        return Webview(self.url)



import secrets
import sqlite3
import uuid
from datetime import datetime, timedelta, timezone
from pathlib import Path

from . import data_models as data_models


# Define the UserPersistence dataclass to handle database operations
class Persistence:
    """
    A class to handle database operations for users and sessions.

    User data is stored in the 'users' table, and session data is stored in the
    'user_sessions' table.

    You can adapt this class to your needs by adding more methods to interact
    with the database or support different databases like MongoDB.

    ## Attributes

    `db_path`: Path to the SQLite database file
    """

    def __init__(self, db_path: Path) -> None:
        """
        Initialize the Persistence instance and ensure necessary tables exist.
        """
        self.conn = sqlite3.connect(db_path)
        self._create_user_table()  # Ensure the users table exists
        self._create_session_table()  # Ensure the sessions table exists

    def _create_user_table(self) -> None:
        """
        Create the 'users' table in the database if it does not exist. The table
        stores user information including id, username, timestamps, and password
        data.
        """
        # Create a cursor object to execute SQL commands
        cursor = self.conn.cursor()

        # We'll store dates as floats (UNIX timestamps), since SQLite doesn't
        # have a native datetime type. Floats are easy to compare and convert,
        # so they work well for this purpose.
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS users (
                id TEXT PRIMARY KEY,
                username TEXT NOT NULL,
                created_at REAL NOT NULL,
                password_hash BLOB NOT NULL,
                password_salt BLOB NOT NULL
            )
        """
        )

        # Commit the changes
        self.conn.commit()

    def _create_session_table(self) -> None:
        """
        Create the 'user_sessions' table in the database if it does not exist.
        The table stores session information including session id, user id, and
        timestamps.
        """
        # Create a cursor object to execute SQL commands
        cursor = self.conn.cursor()

        # Create the user_sessions table if it doesn't exist
        cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS user_sessions (
                id TEXT PRIMARY KEY,
                user_id TEXT NOT NULL,
                created_at REAL NOT NULL,
                valid_until REAL NOT NULL,
                FOREIGN KEY (user_id) REFERENCES users(id)
            )
        """
        )

        # Commit the changes
        self.conn.commit()

    async def create_user(self, user: data_models.AppUser) -> None:
        """
        Add a new user to the database.

        ## Parameters

        `user`: The user object containing user details.
        """
        # Create a cursor object to execute SQL commands
        cursor = self.conn.cursor()

        # SQL command to insert a new user into the table
        cursor.execute(
            """
            INSERT INTO users (id, username, created_at, password_hash, password_salt)
            VALUES (?, ?, ?, ?, ?)
            """,
            (
                str(user.id),
                user.username,
                user.created_at.timestamp(),
                user.password_hash,
                user.password_salt,
            ),
        )

        # Commit the changes
        self.conn.commit()

    async def get_user_by_username(
        self,
        username: str,
    ) -> data_models.AppUser:
        """
        Retrieve a user from the database by username.


        ## Parameters

        `username`: The username of the user to retrieve.


        ## Raises

        `KeyError`: If there is no user with the specified username.
        """
        # Look up the user in the database
        cursor = self.conn.cursor()
        cursor.execute(
            "SELECT * FROM users WHERE username = ? LIMIT 1",
            (username,),
        )

        # Get the first row from the result
        row = cursor.fetchone()

        # If a user was found, wrap it up in a neat Python class
        if row:
            return data_models.AppUser(
                id=uuid.UUID(row[0]),
                username=row[1],
                created_at=datetime.fromtimestamp(row[2], tz=timezone.utc),
                password_hash=row[3],
                password_salt=row[4],
            )

        # If no user was found, signal that with a KeyError
        raise KeyError(username)

    async def get_user_by_id(
        self,
        id: uuid.UUID,
    ) -> data_models.AppUser:
        """
        Retrieve a user from the database by user ID.


        ## Parameters

        `id`: The UUID of the user to retrieve.


        ## Raises

        `KeyError`: If there is no user with the specified ID.
        """
        # Create a cursor object to execute SQL commands
        cursor = self.conn.cursor()

        # SQL command to get user by ID
        cursor.execute(
            "SELECT * FROM users WHERE id = ? LIMIT 1",
            (str(id),),
        )

        # Get the first row from the result
        row = cursor.fetchone()

        # If a user was found, wrap it up in a neat Python class
        if row:
            return data_models.AppUser(
                id=uuid.UUID(row[0]),
                username=row[1],
                created_at=datetime.fromtimestamp(row[2], tz=timezone.utc),
                password_hash=row[3],
                password_salt=row[4],
            )

        # If no user was found, signal that with a KeyError
        raise KeyError(id)

    async def create_session(
        self,
        user_id: uuid.UUID,
    ) -> data_models.UserSession:
        """
        Create a new user session and store it in the database.

        ## Parameters

        `user_id`: The UUID of the user for whom to create the session.
        """
        # Store the time in a variable. It will be used multiple times, and this
        # ensures that all timestamps are consistent.
        now = datetime.now(tz=timezone.utc)

        # Create the new session object
        session = data_models.UserSession(
            id=secrets.token_urlsafe(),
            user_id=user_id,
            created_at=now,
            valid_until=now + timedelta(days=1),
        )

        # Store the session in the database
        cursor = self.conn.cursor()
        cursor.execute(
            """
            INSERT INTO user_sessions (id, user_id, created_at, valid_until)
            VALUES (?, ?, ?, ?)
            """,
            (
                session.id,
                str(session.user_id),
                session.created_at.timestamp(),
                session.valid_until.timestamp(),
            ),
        )
        self.conn.commit()

        # Return the freshly created session
        return session

    async def update_session_duration(
        self,
        session: data_models.UserSession,
        new_valid_until: datetime,
    ) -> None:
        """
        Extend the duration of an existing session. This will update the
        session's validity timestamp both in the given object and the database.

        ## Parameters

        `session`: The session whose duration to extend.

        `new_valid_until`: The new timestamp until which the session should be
            considered valid.
        """
        # Update the session object
        session.valid_until = new_valid_until

        # Commit the changes to persistence
        cursor = self.conn.cursor()

        cursor.execute(
            """
            UPDATE user_sessions
            SET valid_until = ?
            WHERE id = ?
            """,
            (
                session.valid_until.timestamp(),
                session.id,
            ),
        )
        self.conn.commit()

    async def get_session_by_auth_token(
        self,
        auth_token: str,
    ) -> data_models.UserSession:
        """
        Retrieve a user session from the database by authentication token.

        ## Parameters

        `auth_token`: The authentication token (session ID) of the session to
            retrieve.

        ## Raises

        `KeyError`: If there is no session with the specified authentication
        token.
        """
        # Create a cursor object to execute SQL commands
        cursor = self.conn.cursor()

        # Query the database for the session
        cursor.execute(
            "SELECT * FROM user_sessions WHERE id = ? ORDER BY created_at LIMIT 1",
            (auth_token,),
        )

        # Get the first row from the result
        row = cursor.fetchone()

        # If a session was found, wrap it up in a neat Python class
        if row:
            return data_models.UserSession(
                id=row[0],
                user_id=uuid.UUID(row[1]),
                created_at=datetime.fromtimestamp(row[2], tz=timezone.utc),
                valid_until=datetime.fromtimestamp(row[3], tz=timezone.utc),
            )

        # If no session was found, signal that with a KeyError
        raise KeyError(auth_token)



from __future__ import annotations

import dataclasses
import enum
import typing as t

import imy.docstrings
from uniserde import JsonDoc

import rio

from .. import deprecations
from .fundamental_component import FundamentalComponent

__all__ = [
    "MouseEventListener",
    "MouseButton",
    "PressEvent",
    "MouseDownEvent",
    "MouseUpEvent",
    "MouseMoveEvent",
    "MouseEnterEvent",
    "MouseLeaveEvent",
    "DragStartEvent",
    "DragMoveEvent",
    "DragEndEvent",
]


@t.final
class MouseButton(enum.Enum):
    """
    Represents a mouse button.

    This enum represents the different buttons of a computer mouse. It is used
    by some events to inform you which button was pressed or released.
    """

    LEFT = "left"
    MIDDLE = "middle"
    RIGHT = "right"


@dataclasses.dataclass
class _ButtonEvent:
    button: MouseButton


@dataclasses.dataclass
class _PositionedEvent:
    """
    ## Attributes

    `x`: The x coordinate of the mouse when the event was triggered, relative to
        the left side of the window.

    `y`: The y coordinate of the mouse when the event was triggered, relative to
        the top of the window. (So a larger `y` means further down.)
    """

    x: float
    y: float


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class PressEvent(_ButtonEvent, _PositionedEvent):
    """
    Holds information regarding a mouse press event.

    This is a simple dataclass that stores useful information for when the user
    presses a mouse button. You'll typically receive this as argument in
    `on_press` events.

    ## Attributes

    `button`: The mouse button that was pressed.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class MouseDownEvent(_ButtonEvent, _PositionedEvent):
    """
    Holds information regarding a mouse down event.

    This is a simple dataclass that stores useful information for when the user
    presses a mouse button. You'll typically receive this as argument in
    `on_mouse_down` events.

    ## Attributes

    `button`: The mouse button that was pressed.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class MouseUpEvent(_ButtonEvent, _PositionedEvent):
    """
    Holds information regarding a mouse up event.

    This is a simple dataclass that stores useful information for when the user
    releases a mouse button. You'll typically receive this as argument in
    `on_mouse_up` events.

    ## Attributes

    `button`: The mouse button that was released.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
class MouseMoveEvent(_PositionedEvent):
    """
    Holds information regarding a mouse move event.

    This is a simple dataclass that stores useful information for when the user
    moves the mouse. You'll typically receive this as argument in
    `on_mouse_move` events.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
class MouseEnterEvent(_PositionedEvent):
    """
    Holds information regarding a mouse enter event.

    This is a simple dataclass that stores useful information for when the user
    moves the mouse over a component. You'll typically receive this as argument
    in `on_mouse_enter` events.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
class MouseLeaveEvent(_PositionedEvent):
    """
    Holds information regarding a mouse leave event.

    This is a simple dataclass that stores useful information for when the user
    moves the mouse away from a component. You'll typically receive this as
    argument in `on_mouse_leave` events.
    """


@dataclasses.dataclass
class _DragEvent(_ButtonEvent, _PositionedEvent):
    """
    Holds information regarding a drag event.

    This is a simple dataclass that stores useful information for when the user
    drags the mouse. You'll typically receive this as argument in
    `on_drag_start`, `on_drag_move`, and `on_drag_end` events.

    ## Attributes

    `component`: The component located under the mouse cursor when the event
        happened.

    `button`: The mouse button that is held down while dragging.
    """

    component: rio.Component


@t.final
@imy.docstrings.mark_constructor_as_private
class DragStartEvent(_DragEvent):
    """
    Holds information regarding a drag start event.

    This is a simple dataclass that stores useful information for when the user
    starts dragging the mouse. You'll typically receive this as argument in
    `on_drag_start` events.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
class DragMoveEvent(_DragEvent):
    """
    Holds information regarding a drag move event.

    This is a simple dataclass that stores useful information for when the user
    moves the mouse while dragging. You'll typically receive this as argument in
    `on_drag_move` events.
    """


@t.final
@imy.docstrings.mark_constructor_as_private
class DragEndEvent(_DragEvent):
    """
    Holds information regarding a drag end event.

    This is a simple dataclass that stores useful information for when the user
    stops dragging the mouse. You'll typically receive this as argument in
    `on_drag_end` events.
    """


@t.final
class MouseEventListener(FundamentalComponent):
    """
    Old version of `PointerEventListener`

    A newer version of this component exists, called `PointerEventListener`.
    Prefer using that one in new code. `MouseEventListener` will be removed in a
    future Rio release.

    `MouseEventListener` takes a single child component and displays it. It then
    listen for any mouse activity on the child component and reports it through
    its event handlers.


    ## Attributes

    `content`: The child component to display.

    `on_press`: Similar to `on_mouse_up`, but performs additional subtle
        checks, such as that the left mouse button was pressed.

    `on_mouse_down`: Triggered when a mouse button is pressed down while
        the mouse is placed over the child component.

    `on_mouse_up`: Triggered when a mouse button is released while the
        mouse is placed over the child component.

    `on_mouse_move`: Triggered when the mouse is moved while located over
        the child component.

    `on_mouse_enter`: Triggered when the mouse previously was not located
        over the child component, but now is.

    `on_mouse_leave`: Triggered when the mouse previously was located over
        the child component, but now is not.

    `on_drag_start`: Triggered when the user starts dragging the mouse, i.e.
        moving it while holding down a mouse button.

    `on_drag_move`: Triggered when the user moves the mouse while holding down a
        mouse button.

    `on_drag_end`: Triggered when the user stops dragging the mouse.
    """

    content: rio.Component
    _: dataclasses.KW_ONLY
    on_press: rio.EventHandler[PressEvent] = None
    on_mouse_down: rio.EventHandler[MouseDownEvent] = None
    on_mouse_up: rio.EventHandler[MouseUpEvent] = None
    on_mouse_move: rio.EventHandler[MouseMoveEvent] = None
    on_mouse_enter: rio.EventHandler[MouseEnterEvent] = None
    on_mouse_leave: rio.EventHandler[MouseLeaveEvent] = None
    on_drag_start: rio.EventHandler[DragStartEvent] = None
    on_drag_move: rio.EventHandler[DragMoveEvent] = None
    on_drag_end: rio.EventHandler[DragEndEvent] = None

    def __post_init__(self) -> None:
        deprecations.warn(
            since="0.10.5",
            message=f"`MouseEventListener` has been superseded by `PointerEventListener`. Please use `PointerEventListener` instead.",
        )

    def _custom_serialize_(self) -> JsonDoc:
        return {
            "reportPress": self.on_press is not None,
            "reportMouseDown": self.on_mouse_down is not None,
            "reportMouseUp": self.on_mouse_up is not None,
            "reportMouseMove": self.on_mouse_move is not None,
            "reportMouseEnter": self.on_mouse_enter is not None,
            "reportMouseLeave": self.on_mouse_leave is not None,
            "reportDragStart": self.on_drag_start is not None,
            "reportDragMove": self.on_drag_move is not None,
            "reportDragEnd": self.on_drag_end is not None,
        }

    async def _on_message_(self, msg: t.Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg

        msg_type = msg["type"]
        assert isinstance(msg_type, str), msg_type

        # Dispatch the correct event
        if msg_type == "press":
            await self.call_event_handler(
                self.on_press,
                PressEvent(
                    x=msg["x"],
                    y=msg["y"],
                    button=MouseButton(msg["button"]),
                ),
            )

        elif msg_type == "mouseDown":
            await self.call_event_handler(
                self.on_mouse_down,
                MouseDownEvent(
                    x=msg["x"],
                    y=msg["y"],
                    button=MouseButton(msg["button"]),
                ),
            )

        elif msg_type == "mouseUp":
            await self.call_event_handler(
                self.on_mouse_up,
                MouseUpEvent(
                    x=msg["x"],
                    y=msg["y"],
                    button=MouseButton(msg["button"]),
                ),
            )

        elif msg_type == "mouseMove":
            await self.call_event_handler(
                self.on_mouse_move,
                MouseMoveEvent(
                    x=msg["x"],
                    y=msg["y"],
                ),
            )

        elif msg_type == "mouseEnter":
            await self.call_event_handler(
                self.on_mouse_enter,
                MouseEnterEvent(
                    x=msg["x"],
                    y=msg["y"],
                ),
            )

        elif msg_type == "mouseLeave":
            await self.call_event_handler(
                self.on_mouse_leave,
                MouseLeaveEvent(
                    x=msg["x"],
                    y=msg["y"],
                ),
            )

        elif msg_type == "dragStart":
            await self.call_event_handler(
                self.on_drag_start,
                DragStartEvent(
                    button=msg["button"],
                    x=msg["x"],
                    y=msg["y"],
                    component=self.session._weak_components_by_id[
                        msg["component"]
                    ],
                ),
            )

        elif msg_type == "dragMove":
            await self.call_event_handler(
                self.on_drag_move,
                DragMoveEvent(
                    button=msg["button"],
                    x=msg["x"],
                    y=msg["y"],
                    component=self.session._weak_components_by_id[
                        msg["component"]
                    ],
                ),
            )

        elif msg_type == "dragEnd":
            await self.call_event_handler(
                self.on_drag_end,
                DragEndEvent(
                    button=msg["button"],
                    x=msg["x"],
                    y=msg["y"],
                    component=self.session._weak_components_by_id[
                        msg["component"]
                    ],
                ),
            )

        else:
            raise ValueError(
                f"{__class__.__name__} encountered unknown message: {msg}"
            )

        # Refresh the session
        await self.session._refresh()


MouseEventListener._unique_id_ = "MouseEventListener-builtin"



from __future__ import annotations

import typing as t

import rio

from .fundamental_component import FundamentalComponent

__all__ = ["GraphEditor", "GraphStore"]

P = t.ParamSpec("P")


@t.final
class GraphStore:
    """
    ## Metadata

    `public`: False
    """

    def __init__(self) -> None:
        # The graph currently connected to the store. At any point in time there
        # can only be one graph connected to the store.
        #
        # It is the store's responsibility to keep JavaScript and Rio up-to-date
        # with any changes to the graph.
        self._graph: GraphEditor | None = None

        # All node instances currently in the graph
        self._nodes: list[rio.Component] = []

        # All as-of-yet uninstantiated nodes. Since nodes can only be
        # instantiated while in a `build` function, this list will be used to
        # instantiate all nodes when the graph is being (re-) built.
        self._uninstantiated_nodes: list[tuple[t.Callable, tuple, dict]] = []

    def _set_editor(self, editor: GraphEditor) -> list[rio.Component]:
        """
        Connects the graph editor to the store and returns the flat list of
        children of the graph. This list will be shared between the two and
        updated in-place as needed.

        If the store was already previously connected to another graph editor,
        it will disconnect from it first.
        """
        # Already connected to an editor?
        if self._graph is not None:
            self._graph.children = []
            self._graph.session._register_dirty_component(
                self._graph,
                include_children_recursively=False,
            )

        # Connect to the new editor
        self._graph = editor
        return self._nodes

    def add_node(
        self,
        node_type: t.Callable[P, rio.Component],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        """
        Creates a node of the given type and adds it to the graph.
        """
        # Keep track of the node. It can't be instantiated yet, because Rio
        # components can only be created inside of `build` functions.
        self._uninstantiated_nodes.append((node_type, args, kwargs))

        # Tell the graph editor to rebuild
        if self._graph is not None:
            self._graph.session._register_dirty_component(
                self._graph,
                include_children_recursively=False,
            )


@t.final
class GraphEditor(FundamentalComponent):
    """
    ## Metadata

    `public`: False
    """

    children: list[rio.Component]

    def __init__(
        self,
        *children: rio.Component,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.children = list(children)


GraphEditor._unique_id_ = "GraphEditor-builtin"



from __future__ import annotations

import typing as t

import rio

from .. import deprecations
from .fundamental_component import FundamentalComponent

__all__ = [
    "ProgressCircle",
]


@t.final
@deprecations.component_kwarg_renamed(
    since="0.10",
    old_name="size",
    new_name="min_size",
)
class ProgressCircle(FundamentalComponent):
    """
    A progress indicator in the shape of a circle.

    `ProgressCircle` conveys to the user that activity is ongoing. It can either
    display the exact progress as a fraction from 0 to 1, or it can display an
    indeterminate progress animation, which is useful when the exact progress
    isn't known.

    Note that unlike most components in Rio, `ProgressCircle` does not have a
    `natural` size, since the circle can easily be scaled to fit any size.
    Therefore it defaults to a reasonable size which should fit most use cases.

    The linear counterpart to this component is the `ProgressBar`.


    ## Attributes

    `progress`: The progress to display, as a fraction from 0 to 1. If `None`,
        the progress indicator will be indeterminate.

    `color`: The color scheme of the progress indicator. Keeping the default
        is recommended, but it may make sense to change the color in
        case the default is hard to perceive on your background.


    ## Examples

    Here's a minimal example displaying a progress circle that is 40% complete:

    ```python
    rio.ProgressCircle(progress=0.4)
    ```

    You don't have to pass any progress to `rio.ProgressCircle`. If you don't,
    it will indicate to the user that something is happening, without showing
    any specific progress.

    ```python
    rio.ProgressCircle()
    ```
    """

    progress: float | None
    color: rio.ColorSet

    def __init__(
        self,
        progress: float | None = None,
        *,
        color: rio.ColorSet = "keep",
        min_size: float = 3.5,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        """
        ## Parameters

        `progress`: The progress to display, as a fraction from 0 to 1. If
            `None`, the progress indicator will be indeterminate.

        `color`: The color scheme of the progress indicator. Keeping the default
            is recommended, but it may make sense to change the color in case
            the default is hard to perceive on your background.

        `min_size`: The size of the progress indicator. This is equivalent to
            setting a component's `width` and `height` to the same value.

            Note that unlike most components in Rio, `ProgressCircle` does not
            have a `natural` size, since the circle can easily be scaled to fit
            any size. Therefore it defaults to a reasonable size which should
            fit most use cases.
        """
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_size,
            min_height=min_size,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        self.progress = progress
        self.color = color


ProgressCircle._unique_id_ = "ProgressCircle-builtin"



import typing as t

import pytest

import rio.app_server


class FakeComponent(rio.Component):
    pass


class FakeSession:
    def __init__(self) -> None:
        self._base_url = rio.URL("http://unit.test/foo")


@pytest.mark.parametrize(
    "relative_url, expected_result",
    [
        ("", "http://unit.test/foo/bar/hello"),
        ("?param=3#frag", "http://unit.test/foo/bar/hello?param=3#frag"),
        ("page-1?param=3#frag", "http://unit.test/foo/bar/page-1?param=3#frag"),
        ("../page-1?param=3#frag", "http://unit.test/foo/page-1?param=3#frag"),
        ("/page-1?param=3#frag", "http://unit.test/foo/page-1?param=3#frag"),
        ("http://unit.test/?param=3#frag", "http://unit.test/?param=3#frag"),
        (
            "https://some.where/else?param=3#frag",
            "https://some.where/else?param=3#frag",
        ),
    ],
)
def test_make_url_absolute(relative_url: str, expected_result: str) -> None:
    session = t.cast(rio.Session, FakeSession())
    session._base_url = rio.URL("http://unit.test/foo")
    session._active_page_url = rio.URL("http://unit.test/foo/bar/hello")

    result = rio.Session._make_url_absolute(session, relative_url)
    assert str(result) == expected_result


PAGES = [
    rio.ComponentPage(
        name="Home",
        url_segment="",
        build=FakeComponent,
    ),
    rio.ComponentPage(
        name="Page 1",
        url_segment="page-1",
        build=FakeComponent,
    ),
    rio.ComponentPage(
        name="Page 2",
        url_segment="page-2",
        build=FakeComponent,
    ),
    rio.ComponentPage(
        name="Guard to Page 1",
        url_segment="guard-to-page-1",
        build=FakeComponent,
        guard=lambda _: "page-1",
    ),
    rio.ComponentPage(
        name="Guard to Cycle 1",
        url_segment="guard-to-cycle-1",
        build=FakeComponent,
        guard=lambda _: "guard-to-cycle-2",
    ),
    rio.ComponentPage(
        name="Guard to Cycle 2",
        url_segment="guard-to-cycle-2",
        build=FakeComponent,
        guard=lambda _: "guard-to-cycle-1",
    ),
    rio.Redirect(
        url_segment="redirect-to-page-1",
        target="page-1",
    ),
    rio.Redirect(
        url_segment="redirect-to-cycle-1",
        target="guard-to-cycle-2",
    ),
    rio.Redirect(
        url_segment="redirect-to-cycle-2",
        target="guard-to-cycle-1",
    ),
]


@pytest.mark.parametrize(
    "relative_url_before_redirects, relative_url_after_redirects_should",
    [
        # No redirects
        (
            "/page-1",
            "/page-1",
        ),
        (
            "/page-1?foo=bar",
            "/page-1?foo=bar",
        ),
        (
            "/page-1#foo",
            "/page-1#foo",
        ),
        # "Wrong" casing
        (
            "/Page-1",
            "/Page-1",
        ),
        # Redirects by guard
        (
            "/guard-to-page-1",
            "/page-1",
        ),
        (
            "/guard-to-page-1?foo=bar",
            "/page-1?foo=bar",
        ),
        (
            "/guard-to-page-1#foo",
            "/page-1#foo",
        ),
        # Redirects by `rio.Redirect`
        (
            "/redirect-to-page-1",
            "/page-1",
        ),
        (
            "/redirect-to-page-1?foo=bar",
            "/page-1?foo=bar",
        ),
        (
            "/redirect-to-page-1#foo",
            "/page-1#foo",
        ),
        # No such page
        (
            "/non-existent-page",
            "/non-existent-page",
        ),
        (
            "/non-existent-page?foo=bar",
            "/non-existent-page?foo=bar",
        ),
        (
            "/non-existent-page#foo",
            "/non-existent-page#foo",
        ),
        # TODO: Add tests for URLs with Unicode characters & escaped/unescaped
        # URLs
    ],
)
def test_redirects(
    relative_url_before_redirects: str,
    relative_url_after_redirects_should: str,
) -> None:
    """
    Simulate navigation to URLs, run any guards, and make sure the final,
    resulting URL is correct.
    """
    # Create a fake session. It contains everything used by the routing system.
    app = rio.App(pages=PAGES)
    app_server = rio.app_server.TestingServer(app)
    session = app_server.create_dummy_session()

    # Determine the final URL
    active_pages_and_path_arguments, absolute_url_after_redirects_is = (
        rio.routing.check_page_guards(
            session,
            session._base_url.join(rio.URL(relative_url_before_redirects)),
        )
    )

    # Verify the final URL
    absolute_url_after_redirects_should = session._base_url.join(
        rio.URL(relative_url_after_redirects_should)
    )

    assert (
        absolute_url_after_redirects_is == absolute_url_after_redirects_should
    )


def test_url_parameter_parsing_failure() -> None:
    def build_int_page(path_param: int):
        return rio.Text(str(path_param))

    int_page = rio.ComponentPage(
        name="Int Page",
        url_segment="{path_param}",
        build=build_int_page,
    )

    def build_float_page(path_param: float):
        return rio.Text(str(path_param))

    float_page = rio.ComponentPage(
        name="Float Page",
        url_segment="{path_param}",
        build=build_float_page,
    )

    app = rio.App(pages=(int_page, float_page))
    app_server = rio.app_server.TestingServer(app)
    session = app_server.create_dummy_session()

    active_pages_and_path_arguments, _ = rio.routing.check_page_guards(
        session,
        session._base_url.join(rio.URL("/3.5")),
    )

    assert active_pages_and_path_arguments is not None
    assert len(active_pages_and_path_arguments) == 1
    assert active_pages_and_path_arguments[0][0] == float_page
    assert active_pages_and_path_arguments[0][1] == {"path_param": 3.5}


def test_redirect_offsite() -> None:
    """
    Redirect to a site other than this app.
    """
    page = rio.ComponentPage(
        name="Home",
        url_segment="",
        build=FakeComponent,
    )

    app = rio.App(pages=(page,))
    app_server = rio.app_server.TestingServer(app)
    session = app_server.create_dummy_session()

    external_url = rio.URL("http://example.com")

    active_pages_and_path_arguments, absolute_url_after_redirects = (
        rio.routing.check_page_guards(
            session,
            external_url,
        )
    )

    assert active_pages_and_path_arguments is None
    assert absolute_url_after_redirects == external_url



# The "zzz" in the file name is there to ensure that this file runs last, so
# that the remaining tests run without the monkeypatches applied.

import pytest

import rio.testing
from rio.debug.monkeypatches import apply_monkeypatches

apply_monkeypatches()


def test_components_cant_be_instantiated_outside_of_build_methods():
    with pytest.raises(Exception):
        rio.Row()


async def test_type_checking():
    class Comp1(rio.Component):
        attr: list["int"]  # <- partially made of forward references

        def build(self) -> rio.Component:
            return rio.Text("")

    def build():
        rio.Text(
            "foo",
            key="key",
            overflow="wrap",
            margin_x=2,
            grow_x=True,
        )
        rio.Container(rio.Text("bar"))
        Comp1([1, 2, 3])

        return rio.Spacer()

    async with rio.testing.TestClient(build):
        pass


@pytest.mark.parametrize(
    "func_",
    [
        lambda: rio.Text(b"foo"),  # type: ignore
    ],
)
async def test_type_checking_error(func_):
    def build():
        with pytest.raises(TypeError):
            func_()

        return rio.Spacer()

    async with rio.testing.TestClient(build):
        pass


@pytest.mark.parametrize(
    "component_cls",
    [
        rio.Spacer,
        rio.Row,
        rio.Grid,
    ],
)
async def test_component_class_can_be_used_as_build_function(
    component_cls: type[rio.Component],
):
    def build():
        return rio.PageView(fallback_build=component_cls)

    async with rio.testing.TestClient(build):
        pass


async def test_init_cannot_read_state_properties():
    # Accessing state properties in `__init__` is not allowed because state
    # bindings aren't initialized yet at that point. In development mode, trying
    # to access a state property in `__init__` should raise an exception.
    init_executed = False
    accessing_foo_raised_exception = accessing_margin_top_raised_exception = (
        False
    )

    class IllegalComponent(rio.Component):
        foo: int

        def __init__(self, foo: int):
            super().__init__()

            self.foo = foo

            nonlocal accessing_foo_raised_exception
            try:
                _ = self.foo
            except Exception:
                accessing_foo_raised_exception = True

            nonlocal accessing_margin_top_raised_exception
            try:
                _ = self.margin_top
            except Exception:
                accessing_margin_top_raised_exception = True

            nonlocal init_executed
            init_executed = True

        def build(self) -> rio.Component:
            return rio.Text("hi", margin_top=self.margin_top)

    class Container(rio.Component):
        def build(self) -> rio.Component:
            return IllegalComponent(17)

    async with rio.testing.TestClient(Container):
        assert init_executed
        assert accessing_foo_raised_exception
        assert accessing_margin_top_raised_exception



import rio


# <component>
class ContentCard(rio.Component):
    """
    A component that represents a card with a header and content.


    ## Attributes:

    `header`: The text to display as the card's header.

    `content`: The content component to display within the card.
    """

    header: str
    content: rio.Component

    def build(self) -> rio.Component:
        return rio.Card(
            content=rio.Column(
                # Add header with bold font
                rio.Text(
                    self.header,
                    style=rio.TextStyle(
                        font_size=1.5,
                        font_weight="bold",
                    ),
                ),
                # Add content
                self.content,
                # Fill up remaining space with a spacer, therefore all cards
                # have the same height
                rio.Spacer(),
                spacing=1,
                margin=1,
            ),
        )


# </component>



from __future__ import annotations

import abc
import dataclasses
import typing as t

import typing_extensions as te
from uniserde import Jsonable

import rio

from . import assets, deprecations, utils
from .color import Color
from .self_serializing import SelfSerializing
from .utils import ImageLike

__all__ = [
    "Fill",
    "ImageFill",
    "LinearGradientFill",
    "RadialGradientFill",
    "SolidFill",
    "FrostedGlassFill",
]


@deprecations.deprecated(
    since="0.8.5",
    description="The `Fill` base class will be removed.",
)
class Fill(SelfSerializing, abc.ABC):
    """
    Base class for how shapes are filled.

    This is a base class for all fills. Fills determine how the inside of a
    shape is colored.

    This class is abstract and cannot be instantiated directly. Instead, use one
    of its subclasses:

    - `SolidFill`
    - `ImageFill`
    - `LinearGradientFill`
    - `FrostedGlassFill`
    """


@dataclasses.dataclass(frozen=True, eq=True)
class SolidFill(Fill):
    """
    Fills a shape with a single color.

    `SolidFill` is the simplest of all fills. It fills the entire shape with a
    single, solid color.

    ## Attributes

    `color`: The color to fill the shape with.
    """

    color: Color

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return {
            "type": "solid",
            "color": self.color.srgba,
        }


@dataclasses.dataclass(frozen=True, eq=True)
class LinearGradientFill(Fill):
    """
    Fills a shape with a linear gradient.

    `LinearGradientFill` fills the shape with a linear gradient. The gradient
    can have any number of stops, each with a color and a position. The gradient
    will smoothly transition between the colors at the given positions. The
    positions are given as are given as fractions, where 0 is the start of the
    gradient and 1 is the end.


    ## Attributes

    `stops`: The different colors that comprise the gradient, along with where
        they are positioned.

        The stops are given as tuples. Each tuple contains a color and a
        position. The position is a fraction, where 0 is the start of the
        gradient and 1 is the end.

        The order of the stops has no effect.

        There must be at least one stop.

    `angle_degrees`: The angle of the gradient, in degrees. 0 degrees points
        straight to the right, and the angle increases counterclockwise.
        (This lines up with how angles are typically used mathematically.)
    """

    stops: tuple[tuple[Color, float], ...]
    angle_degrees: float = 0.0

    def __init__(
        self,
        *stops: rio.Color | tuple[rio.Color, float],
        angle_degrees: float = 0.0,
    ) -> None:
        # Postprocess & store the stops
        vars(self).update(
            stops=utils.verify_and_interpolate_gradient_stops(stops),
            angle_degrees=angle_degrees,
        )

    def _as_css_background(self, sess: rio.Session) -> str:
        # Special case: Just one color
        if len(self.stops) == 1:
            return f"#{self.stops[0][0].hexa}"

        # Proper gradient
        stop_strings = []

        for stop in self.stops:
            color = stop[0]
            position = stop[1]
            stop_strings.append(f"#{color.hexa} {position * 100}%")

        return f"linear-gradient({90 - self.angle_degrees}deg, {', '.join(stop_strings)})"

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return {
            "type": "linearGradient",
            "stops": [
                (color.srgba, position) for color, position in self.stops
            ],
            "angleDegrees": self.angle_degrees,
        }


class ImageFill(Fill):
    """
    Fills a shape with an image.

    `ImageFill` fills the shape's background with an image.

    The image can be scaled to fit the shape in one of three ways:

    - `fit`: The image is scaled to fit entirely inside the shape, while
      maintaining its aspect ratio. This is the default.
    - `stretch`: The image is stretched to fill the shape, distorting it if
      necessary.
    - `zoom`: The image is scaled to fill the shape entirely, while maintaining
      its aspect ratio. This may cause the image to overflow the shape.
    """

    def __init__(
        self,
        image: ImageLike,
        *,
        fill_mode: t.Literal["fit", "stretch", "zoom", "tile"] = "fit",
        tile_size: tuple[float, float] = (10, 10),
    ) -> None:
        """
        ## Parameters

        `image`: The image to fill the shape with. fill_mode: How the image
            should be scaled to fit the shape.

        `fill_mode`: How the image should be scaled to fit the shape. If `fit`,
            the image is scaled to fit entirely inside the shape. If `stretch`,
            the image is stretched to fill the shape exactly, possibly
            distorting it in the process. If `zoom`, the image is scaled to fill
            the shape entirely, possibly overflowing. If `tile`, the image is
            repeated to fill the shape.

        `tile_size`: The width and height of the tile in font heights. This is
            ignored if `fill_mode` is not `"tile"`.

        """
        self._image_asset = assets.Asset.from_image(image)
        self._fill_mode = fill_mode
        self._tile_size = tile_size

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return {
            "type": "image",
            "imageUrl": self._image_asset._serialize(sess),
            "fillMode": self._fill_mode,
            "tileSize": self._tile_size,
        }

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ImageFill):
            return NotImplemented

        return (
            self._image_asset == other._image_asset
            and self._fill_mode == other._fill_mode
        )

    def __hash__(self) -> int:
        return hash((self._image_asset, self._fill_mode))

    def _as_css_background(self, sess: rio.Session) -> str:
        # Fetch the escaped URL. That way it cannot interfere with the CSS syntax
        image_url = self._image_asset._serialize(sess)
        css_url = f"url('{image_url}')"

        if self._fill_mode == "fit":
            return f"{css_url} center/contain no-repeat"
        elif self._fill_mode == "stretch":
            return f"{css_url} top left / 100% 100%"
        elif self._fill_mode == "zoom":
            return f"{css_url} center/cover no-repeat"
        elif self._fill_mode == "tile":
            return f"{css_url} top left / {self._tile_size[0]}rem {self._tile_size[1]}rem repeat"
        else:
            # Invalid fill mode
            raise Exception(
                f"Invalid fill mode for image fill: {self._fill_mode}"
            )


@dataclasses.dataclass(frozen=True, eq=True)
class RadialGradientFill(Fill):
    """
    Fills a shape with a circular gradient.

    `RadialGradientFill` fills the shape with a circular gradient that emanates
    from a center point. The gradient can have any number of stops, each with a
    color and a position. The gradient will smoothly transition between the
    colors at the given positions. The positions are given as fractions, where 0
    is the center of the gradient and 1 is the edge.

    ## Attributes

    `stops`: The different colors that comprise the gradient, along with where
        they are positioned.

        The stops are given as tuples. Each tuple contains a color and a
        position. The position is a fraction, where 0 is the center of the
        gradient and 1 is the edge.

        The order of the stops has no effect.

        There must be at least one stop.

    `center_x`: The x-coordinate of the center of the gradient, as a fraction
        of the shape's width. 0.5 is the center of the shape.

    `center_y`: The y-coordinate of the center of the gradient, as a fraction
        of the shape's height. 0.5 is the center of the shape.
    """

    stops: tuple[tuple[Color, float], ...]

    center_x: float = 0.5
    center_y: float = 0.5

    def __init__(
        self,
        *stops: rio.Color | tuple[rio.Color, float],
        center_x: float = 0.5,
        center_y: float = 0.5,
    ) -> None:
        # Postprocess & store the stops
        vars(self).update(
            stops=utils.verify_and_interpolate_gradient_stops(stops),
            center_x=center_x,
            center_y=center_y,
        )

    def _as_css_background(self, sess: rio.Session) -> str:
        # Special case: Just one color
        if len(self.stops) == 1:
            return f"#{self.stops[0][0].hexa}"

        # Proper gradient
        stop_strings = []

        for stop in self.stops:
            color = stop[0]
            position = stop[1]
            stop_strings.append(f"#{color.hexa} {position * 100}%")

        center_position = f"{self.center_x * 100}% {self.center_y * 100}%"
        return f"radial-gradient(circle at {center_position}, {', '.join(stop_strings)})"

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return {
            "type": "radialGradient",
            "stops": [
                (color.srgba, position) for color, position in self.stops
            ],
            "centerX": self.center_x,
            "centerY": self.center_y,
        }


@dataclasses.dataclass(frozen=True, eq=True)
class FrostedGlassFill(Fill):
    """
    Fills a shape with a frosted glass effect.

    `FrostedGlassFill` fills the shape with a color and applies a blur effect to
    the background, creating a frosted glass appearance.

    Make sure to use a color with transparency, otherwise it will look like a
    `SolidFill`.

    ## Attributes

    `color`: The color to fill the shape with.
    `blur_size`: The amount of blur applied to the fill.
    """

    color: Color
    blur_size: float = 0.3

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return {
            "type": "frostedGlass",
            "color": self.color._serialize(sess),
            "blurSize": self.blur_size,
        }


_FillLike: te.TypeAlias = (
    SolidFill
    | LinearGradientFill
    | RadialGradientFill
    | ImageFill
    | FrostedGlassFill
    | Color
)



# <additional-imports>
import rio

from .. import data_models, theme

# </additional-imports>


# <component>
class Faq(rio.Component):
    """
    A responsive FAQ component that toggles the visibility of an answer.

    The component includes:
    - A header that serves as a clickable question.
    - A body that displays the answer when expanded.
    - An open/close icon to indicate the state.


    ## Attributes:

    `header`: The question or title of the FAQ item.

    `body`: The answer or description associated with the FAQ.

    `is_last`: Indicates if this is the last FAQ item in a list for additional
    spacing.

    `is_open`: Determines whether the FAQ item is expanded or collapsed.
    """

    header: str
    body: str
    is_last: bool = False
    is_open: bool = True

    def _on_toggle(self, _: rio.PointerEvent) -> None:
        """
        Toggles the `is_open` state when the FAQ header is clicked.
        """
        self.is_open = not self.is_open

    def build(self) -> rio.Component:
        # Get the current device from the session
        device = self.session[data_models.PageLayout].device
        # Specify margin_y to add space between header and body
        margin_y = 1.8

        # Define the icon and content based on the open state
        if self.is_open:
            icon = rio.Icon(
                "material/keyboard_arrow_down",  # Down arrow when expanded
                min_height=1.5,
                min_width=1.5,
                align_x=1,  # Align icon to the right
            )
            # Column to hold answer and separators
            content = rio.Column()
            # Add a separator above the body
            content.add(
                rio.Separator(
                    color=theme.NEUTRAL_COLOR_BRIGHTER,
                ),
            )
            # Add the body text
            content.add(
                rio.Container(
                    rio.Text(
                        self.body,
                        style=theme.DARKER_TEXT,
                        overflow="wrap",
                        margin_y=margin_y,
                    ),
                ),
            )
            # Add a separator if this is the last FAQ for better spacing
            if self.is_last:
                content.add(
                    rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                )

        else:
            icon = rio.Icon(
                "material/chevron_right",  # Right arrow when collapsed
                min_height=1.5,
                min_width=1.5,
                align_x=1,  # Align icon to the right
            )
            # No content displayed when collapsed
            content = None

        # Construct the header row with pointer interaction
        return rio.Column(
            rio.PointerEventListener(
                rio.Rectangle(
                    content=rio.Row(
                        rio.Text(
                            self.header,
                            style=rio.TextStyle(
                                font_weight="bold",
                                # Adjust font size for device
                                font_size=1.2 if device == "desktop" else 1.1,
                            ),
                            selectable=False,  # Prevent text selection
                        ),
                        icon,  # Add the toggle icon
                        margin_y=margin_y,  # Add vertical margin
                    ),
                    fill=rio.Color.TRANSPARENT,
                    # Set pointer cursor for interactivity
                    cursor=rio.CursorStyle.POINTER,
                ),
                on_press=self._on_toggle,
            ),
            # Add a switcher to toggle the visibility of the content
            rio.Switcher(
                content=content,
                transition_time=0.2,
            ),
        )


# </component>



class RioDeprecationWarning(DeprecationWarning):
    """
    The user used functionality that has been deprecated.
    """


class RioPotentialMistakeWarning(Warning):
    """
    The user did something sketchy that's probably a bad idea.
    """



from __future__ import annotations

# <additional-imports>
import rio

# </additional-imports>


# <component>
class SubscribeInputButton(rio.Component):
    """
    A component for collecting email subscriptions.
    """

    def build(self) -> rio.Component:
        """
        Constructs the subscription input and button component.

        The component consists of:
        - A pill-styled text input field for email input.
        - A rounded "Subscribe" button with hover and ripple effects.
        """
        return rio.Stack(
            # Email input field
            rio.TextInput(
                "",
                label="Enter your email address",
                style="pill",
                min_height=3,
                align_y=0,
            ),
            # Subscribe button
            rio.Rectangle(
                content=rio.Text(
                    "Subscribe",
                    style=rio.TextStyle(
                        font_size=0.75,
                        fill=self.session.theme.background_color,
                    ),
                    selectable=False,  # Make the text non-selectable
                    margin=0.5,
                ),
                fill=self.session.theme.primary_color,
                hover_fill=self.session.theme.primary_color.darker(0.2),
                transition_time=0.1,  # Smooth transition for hover effects
                corner_radius=99999,  # Fully rounded corners for the button
                # Add a pointer cursor and ripple effect on hover
                cursor=rio.CursorStyle.POINTER,
                ripple=True,
                # Align and add margin
                align_x=1,
                align_y=0.5,
                margin_right=0.5,
            ),
        )


# </component>



"""
This file reads all bootstrap icons from their GitHub repository and packs them
into a `.tar.xz` archive that can be used by Rio as icon set.

The repository is expected to be available locally already - this script does
not clone it.
"""

import tarfile
import tempfile
import typing as t
from pathlib import Path
from xml.etree import ElementTree as ET

import revel


def icon_name_from_path(path: Path) -> tuple[str, str | None | None]:
    """
    Configure: Given the path to the icon, return the icon's name and
    optionally variant.

    If `None` is returned the file is skipped.
    """
    # Normalize the name
    name = path.stem
    name = name.replace("-", "_")
    assert all(c.isalnum() or c == "_" for c in name), path

    # See if this is a variant
    known_variants = [
        "fill",
    ]

    for variant in known_variants:
        if name.endswith(f"_{variant}"):
            name = name.removesuffix(f"_{variant}")
            break
    else:
        variant = None

    return name, variant


def build_icon_set(
    *,
    input_files: t.Iterable[Path],
    output_path: Path,
    set_name: str,
    limit: int | None = None,
) -> None:
    """
    Helper function for building Rio compatible icon sets.

    This function reads a list of SVG files, postprocesses them to fit Rio's
    requirements and dumps them into a compressed archive, again matching Rio's
    expectations.

    Depending on how exactly your input SVGs are structured, this may not be
    enough. Feel free to adapt the function to your exact files.

    ## Parameters

    `input_files`: An iterable over the paths to the SVG files to process.
        (`Path.glob` is your friend here)

    `output_path`: The path to the `.tar.xz` archive to write the icon set to.

    `set_name`: The name of the icon set. This will be used to access the icons.

    `limit`: If set to a number, only process this many files. Set to `None` to
        process all files (Useful for debugging).
    """

    # Find all files to process
    in_file_paths = list(input_files)
    revel.print(f"Found {len(in_file_paths)} file(s)")

    # Enforce an order so that the output is deterministic
    in_file_paths.sort()

    # Apply the limit
    if limit is not None:
        in_file_paths = in_file_paths[:limit]
        revel.warning(
            f"Only processing the first {limit} file(s) (set `limit` to `None` to disable)"
        )

    # Process all files and store the results in a temporary directory
    revel.print_chapter("Processing files")
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)

        with revel.ProgressBar(max=len(in_file_paths), unit="count") as bar:
            for ii, file_path in enumerate(in_file_paths):
                bar.progress = ii

                # Extract the name and variant of the icon. If this function returns
                # `None` the file is skipped.
                parsed = icon_name_from_path(file_path.resolve())

                if parsed is None:
                    revel.print(f"{file_path.name} -> [bold]skipped[/bold]")
                    continue

                icon_name, icon_variant = parsed
                variant_suffix = (
                    "" if icon_variant is None else f":{icon_variant}"
                )

                revel.print(
                    f"{file_path.name} -> {set_name}/{icon_name}{variant_suffix}"
                )

                # Parse the SVG
                svg_str = file_path.read_text()
                tree = ET.fromstring(svg_str)

                # Strip problematic attributes
                if "width" in tree.attrib:
                    del tree.attrib["width"]

                if "height" in tree.attrib:
                    del tree.attrib["height"]

                if "fill" in tree.attrib:
                    del tree.attrib["fill"]

                if "class" in tree.attrib:
                    del tree.attrib["class"]

                # Determine the output path for this icon
                icon_out_path = tmp_dir

                if icon_variant is not None:
                    icon_out_path /= icon_variant

                icon_out_path /= f"{icon_name}.svg"

                # Suppress weird "ns0:" prefixes everywhere
                ET.register_namespace("", "http://www.w3.org/2000/svg")

                # Write the SVG
                icon_out_path.parent.mkdir(parents=True, exist_ok=True)

                with open(icon_out_path, "w") as f:
                    f.write(
                        ET.tostring(
                            tree,
                            encoding="unicode",
                            default_namespace=None,
                        )
                    )

        # Compress the temporary directory
        revel.print_chapter("Compressing files")

        with tarfile.open(output_path, "w:xz") as out_file:
            out_file.add(tmp_dir, arcname=set_name)

    revel.print_chapter(None)
    revel.success(
        f"[bold]Done![/] You can find the result at [bold]{output_path}[/bold]"
    )


if __name__ == "__main__":
    PROJECT_DIR = Path(__file__).resolve().parent.parent

    # Clone the material-icons repoisitory from GitHub:
    #
    # https://github.com/marella/material-symbols
    # build_icon_set(
    #     input_files=(
    #         PROJECT_DIR
    #         / "thirdparty"
    #         / "material-icons"
    #         / "thirdparty"
    #         / "material-symbols"
    #         / "svg"
    #         / "500"
    #         / "rounded"
    #     ).glob("*.svg"),
    #     output_path=PROJECT_DIR / "material.tar.xz",
    #     set_name="material",
    #     limit=None,
    # )

    # Clone the bootstrap-icons repository from GitHub:
    #
    # https://github.com/twbs/icons
    build_icon_set(
        input_files=(PROJECT_DIR / "thirdparty" / "bootstrap" / "icons").glob(
            "*.svg"
        ),
        output_path=PROJECT_DIR / "bootstrap.tar.xz",
        set_name="bootstrap",
        limit=None,
    )



import rio


# <component>
class HamburgerButton(rio.Component):
    """
    A toggleable hamburger button component.

    This component displays a hamburger menu icon when the menu is closed and a
    close icon when the menu is open. It allows toggling the `is_open` state
    through a pointer (click/touch) event.


    ## Attributes:

    `is_open`: Indicates whether the menu is currently open or closed.
    """

    is_open: bool = False

    def _on_toggle_open(self, _: rio.PointerEvent) -> None:
        """
        Handles the toggle action when the button is pressed.

        This method toggles the `is_open` state between `True` and `False`.
        """
        self.is_open = not self.is_open

    def build(self) -> rio.Component:
        """
        Builds the hamburger button component.

        The button dynamically changes its icon and appearance based on the
        `is_open` state.
        """
        # Determine the icon and fill color based on the open state
        if self.is_open:
            # Display a close icon when the menu is open
            icon = "material/close"
            fill = self.session.theme.text_style.fill
        else:
            # Display a hamburger menu icon when closed
            icon = "material/menu"
            fill = self.session.theme.text_style.fill

        # Ensure the fill is a valid rio.Color instance
        assert isinstance(fill, rio.Color)

        return rio.PointerEventListener(
            rio.Rectangle(
                content=rio.Icon(
                    icon=icon,
                    fill=fill,
                    min_width=2.2,
                    min_height=2.2,
                ),
                fill=rio.Color.TRANSPARENT,
                # Show pointer cursor on hover
                cursor=rio.CursorStyle.POINTER,
            ),
            on_press=self._on_toggle_open,
        )


# </component>



import os
import typing as t
from pathlib import Path

import rio.components.class_container
import rio.debug.dev_tools.dev_tools_connector

from . import (
    deploy_page,
    docs_page,
    icons_page,
    project_page,
    rio_developer_page,
    theme_picker_page,
    tree_page,
)


class DevToolsSidebar(rio.Component):
    show_rio_developer_page: bool = False

    selected_page: (
        t.Literal[
            "project",
            "tree",
            "docs",
            "deploy",
            "rio-developer",
        ]
        | None
    ) = None

    def __post_init__(self) -> None:
        # Decide whether to show the Rio Developer page. Ideally, the page would
        # be displayed to developers of Rio itself - without them having to do
        # anything - while not showing it to developers _using_ Rio.
        #
        # One easy way to achieve this is to look whether a `pyproject.toml`
        # file is located right outside of Rio's root directory.
        rio_root_directory = Path(rio.__file__).parent
        pyproject_toml_path = rio_root_directory.parent / "pyproject.toml"
        self.show_rio_developer_page = pyproject_toml_path.exists()

        # HOWEVER, always allow forcing the tools to be visible via an
        # environment variable:
        try:
            os.environ["RIO_DEV"]
        except KeyError:
            pass
        else:
            self.show_rio_developer_page = True

    def get_selected_page(self) -> rio.Component | None:
        REGULAR_PAGE_WIDTH = 22
        WIDE_PAGE_WIDTH = 34

        # Nothing selected
        if self.selected_page is None:
            return None

        # Project
        if self.selected_page == "project":
            return project_page.ProjectPage(
                min_width=REGULAR_PAGE_WIDTH,
            )

        # Tree
        if self.selected_page == "tree":
            return tree_page.TreePage(
                min_width=WIDE_PAGE_WIDTH,
            )

        # Icons
        if self.selected_page == "icons":
            return icons_page.IconsPage(
                min_width=WIDE_PAGE_WIDTH,
            )

        # Theme
        if self.selected_page == "theme":
            return theme_picker_page.ThemePickerPage(
                min_width=WIDE_PAGE_WIDTH,
            )

        # Docs
        if self.selected_page == "docs":
            return docs_page.DocsPage(
                min_width=REGULAR_PAGE_WIDTH,
            )

        # Deploy
        if self.selected_page == "deploy":
            return deploy_page.DeployPage(
                min_width=REGULAR_PAGE_WIDTH,
            )

        # Rio Developer
        if self.selected_page == "rio-developer":
            return rio_developer_page.RioDeveloperPage(
                min_width=REGULAR_PAGE_WIDTH,
            )

        # Anything else / TODO
        return rio.Text(
            f"TODO: {self.selected_page}",
            justify="center",
            margin=2,
            min_width=REGULAR_PAGE_WIDTH,
        )

    def build(self) -> rio.Component:
        names = [
            # "Project",
            "Tree",
            "Icons",
            "Theme",
            # "Docs",
            "Deploy",
        ]

        icons = [
            # "rio/logo",
            "material/view_quilt",
            "material/emoji_people",
            "material/palette",
            # "material/library_books",
            "material/rocket_launch",
        ]

        values = [
            # "project",
            "tree",
            "icons",
            "theme",
            # "docs",
            "deploy",
        ]

        # If developing Rio itself, show the Rio Developer page
        if self.show_rio_developer_page:
            names.append("Rio Dev")
            icons.append("rio/logo")
            values.append("rio-developer")

        return rio.Rectangle(
            # Make sure everything has a background, otherwise the component
            # highlighter will be visible behind this component
            fill=self.session.theme.background_color,
            content=rio.Row(
                # Big fat line to separate the dev tools from the rest of the page
                rio.Rectangle(
                    min_width=0.3,
                    fill=self.session.theme.primary_palette.background,
                ),
                # Currently active page
                rio.components.class_container.ClassContainer(
                    rio.Switcher(self.get_selected_page()),
                    classes=[
                        "rio-switcheroo-neutral",
                        "rio-dev-tools-background",
                    ],
                ),
                # Navigation
                rio.Column(
                    rio.SwitcherBar(
                        names=names,
                        icons=icons,
                        values=values,
                        allow_none=True,
                        orientation="vertical",
                        spacing=2,
                        color="primary",
                        selected_value=self.bind().selected_page,
                        margin=0.2,
                    ),
                    rio.Spacer(),
                    rio.Tooltip(
                        anchor=rio.Link(
                            "Help",
                            target_url="https://rio.dev/docs/howto/devtools",
                            open_in_new_tab=True,
                            margin_y=1,
                            align_x=0.5,
                        ),
                        tip=rio.Markdown(
                            "This is rio's dev tools sidebar. It automatically disappears if you run your app with the `--release` flag. Click for more information.",
                            # Tooltips use the `min-content` for their width,
                            # which is unreadable. Set a width that looks more
                            # natural.
                            min_width=min(17, self.session.window_width - 1),
                        ),
                        position="left",
                    ),
                    rio.debug.dev_tools.dev_tools_connector.DevToolsConnector(),
                ),
            ),
        )



from __future__ import annotations

import typing as t

from uniserde import JsonDoc

import rio

from .. import icon_registry
from .fundamental_component import FundamentalComponent

__all__ = [
    "Link",
]


@t.final
class Link(FundamentalComponent):
    """
    Navigates to a page or URL when pressed.

    `Link`s display a short text, or an arbitrary component, and navigate to a
    page or URL when clicked.


    ## Attributes

    `child_text`: The text to display inside the link.

    `icon`: Optionally an icon to display next to the text. This is only visible
        if the content is a string.

    `child_component`: The component to display inside the link.

    `target_url`: The page or URL to navigate to when clicked.

    `open_in_new_tab`: Whether to open the link in a new tab. Defaults to
        `False`.


    ## Examples

    This minimal example will display a simple text and navigate away when
    pressed.

    ```python
    rio.Link("Click me!", "https://example.com")
    ```

    You can also use a component as the child of the link:

    ```python
    rio.Link(
        rio.Button("Click me!"),
        "https://example.com",
    )
    ```
    """

    # Exactly one of these will be set, the other is `None`
    child_text: str | None
    child_component: rio.Component | None
    target_url: rio.URL | str
    open_in_new_tab: bool
    icon: str | None

    # The serializer can't handle Union types. Override the constructor, so it
    # splits the child into two values
    def __init__(
        self,
        content: rio.Component | str,
        target_url: rio.URL | str,
        *,
        icon: str | None = None,
        open_in_new_tab: bool = False,
        key: str | int | None = None,
        margin: float | None = None,
        margin_x: float | None = None,
        margin_y: float | None = None,
        margin_left: float | None = None,
        margin_top: float | None = None,
        margin_right: float | None = None,
        margin_bottom: float | None = None,
        min_width: float = 0,
        min_height: float = 0,
        # MAX-SIZE-BRANCH max_width: float | None = None,
        # MAX-SIZE-BRANCH max_height: float | None = None,
        grow_x: bool = False,
        grow_y: bool = False,
        align_x: float | None = None,
        align_y: float | None = None,
        # SCROLLING-REWORK scroll_x: t.Literal["never", "auto", "always"] = "never",
        # SCROLLING-REWORK scroll_y: t.Literal["never", "auto", "always"] = "never",
    ) -> None:
        """
        ## Parameters

        `content`: The text or component to display inside the link.

        `target_url`: The page or URL to navigate to when clicked.
        """
        super().__init__(
            key=key,
            margin=margin,
            margin_x=margin_x,
            margin_y=margin_y,
            margin_left=margin_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            min_width=min_width,
            min_height=min_height,
            # MAX-SIZE-BRANCH max_width=max_width,
            # MAX-SIZE-BRANCH max_height=max_height,
            grow_x=grow_x,
            grow_y=grow_y,
            align_x=align_x,
            align_y=align_y,
            # SCROLLING-REWORK scroll_x=scroll_x,
            # SCROLLING-REWORK scroll_y=scroll_y,
        )

        if isinstance(content, str):
            self.child_text = content
            self.child_component = None
        else:
            self.child_text = None
            self.child_component = content

        self.target_url = target_url
        self.open_in_new_tab = open_in_new_tab
        self.icon = icon

        self._properties_set_by_creator_.update(
            ("child_text", "child_component")
        )

    def __post_init__(self) -> None:
        # Verify that the icon exists. This makes sure any crashes happen
        # immediately, rather than during the next refresh.
        if self.icon is not None:
            icon_registry.get_icon_svg(self.icon)

    def _custom_serialize_(self) -> JsonDoc:
        # Get the full URL to navigate to
        target_url_absolute = self.session._make_url_absolute(self.target_url)

        # Serialize everything
        return {
            "targetUrl": str(target_url_absolute),
        }


Link._unique_id_ = "Link-builtin"



from __future__ import annotations

import typing as t

import rio


# <code>
class Field(rio.Component):
    value: t.Literal["X", "O", ""]
    dim: bool

    on_press: rio.EventHandler[[]] = None

    def build(self) -> rio.Component:
        # If the field is empty, allow the player to press on it. Since buttons
        # would look out of place here, cards are a nice alternative.
        if self.value == "":
            return rio.Card(
                content=rio.Spacer(
                    min_width=3,
                    min_height=3,
                ),
                on_press=self.on_press,
            )

        # For fields that already contain an X or O, show the respective icon.
        # Also vary the color based on the player.
        if self.value == "X":
            color = rio.Color.RED
            icon = "material/close"
        else:
            color = rio.Color.BLUE
            icon = "material/circle"

        # If a player has won, and this field isn't part of the winning
        # combination, dim it.
        if self.dim:
            color = color.replace(opacity=0.2)

        return rio.Icon(
            icon=icon,
            fill=color,
            min_width=3,
            min_height=3,
        )


# </code>



import rio
from tests.utils.layouting import verify_layout


async def test_stack() -> None:
    """
    All children in stacks should be the same size.
    """
    layouter = await verify_layout(
        lambda: rio.Stack(
            rio.Text("Small", key="small_text", min_width=10, min_height=20),
            rio.Text("Large", key="large_text", min_width=30, min_height=40),
            align_x=0,
            align_y=0,
        )
    )

    small_layout = layouter.get_layout_by_key("small_text")

    assert small_layout.left_in_viewport_inner == 0
    assert small_layout.top_in_viewport_inner == 0

    assert small_layout.allocated_inner_width == 30
    assert small_layout.allocated_inner_height == 40

    large_layout = layouter.get_layout_by_key("large_text")

    assert large_layout.left_in_viewport_inner == 0
    assert large_layout.top_in_viewport_inner == 0

    assert large_layout.allocated_inner_width == 30
    assert large_layout.allocated_inner_height == 40



from __future__ import annotations

import rio

from .. import components as comps


# <component>
class RootComponent(rio.Component):
    """
    This component will be used as the root for the app. This means that it will
    always be visible, regardless of which page is currently active.

    This makes it the perfect place to put components that should be visible on
    all pages, such as a navbar or a footer.

    Additionally, the root will contain a `rio.PageView`. Page views don't have
    any appearance of their own, but they are used to display the content of the
    currently active page. Thus, we'll always see the navbar and footer, with
    the content of the current page sandwiched in between.
    """

    def build(self) -> rio.Component:
        return rio.Column(
            # The navbar contains a `rio.Overlay`, so it will always be on top
            # of all other components.
            comps.Navbar(),
            # The page view will display the content of the current page.
            rio.PageView(
                # Make sure the page view takes up all available space. Without
                # this the navbar would be assigned the same space as the page
                # content.
                grow_y=True,
            ),
        )


# </component>



"""
Tables support numpy-style 2D indexing. This is rather complex, hence the
tests here.
"""

import typing as t

import pytest

import rio


# Helper class for easily creating indices
class MakeIndex:
    def __getitem__(self, index) -> t.Any:
        return index


make_index = MakeIndex()


@pytest.mark.parametrize(
    "index,enable_column_names,result_should",
    [
        # Valid index without anything special
        (
            make_index[1:3, 4:7],
            False,
            (4, 1, 3, 2),
        ),
        # Selecting an entire column
        (
            make_index["a"],
            True,
            (0, 0, 1, 20),
        ),
        (
            make_index["b"],
            True,
            (1, 0, 1, 20),
        ),
        # Attempting to select a column, but no column names are available
        (
            make_index["a"],
            False,
            ValueError,
        ),
        # Selecting a non-existent column
        (
            make_index["invalid_column_name"],
            True,
            KeyError,
        ),
        # Selecting some rows in a column
        (
            make_index[1:3, "a"],
            True,
            (0, 1, 1, 2),
        ),
        # Attempting to select by name in the wrong axis
        (
            make_index["a", 0],
            True,
            ValueError,
        ),
        (
            make_index["a", "b"],
            True,
            ValueError,
        ),
        # Wrong number of indices
        (
            make_index[0],
            True,
            ValueError,
        ),
        (
            make_index[1, 2, 3],
            True,
            ValueError,
        ),
        (
            make_index[1, 2, 3, 4],
            True,
            ValueError,
        ),
        # Wrong datatypes
        (
            make_index[1.0],
            False,
            ValueError,
        ),
        (
            make_index[1.0, 2.0],
            False,
            ValueError,
        ),
        (
            make_index[1.0, 2.0, 3.0],
            False,
            ValueError,
        ),
        (
            make_index[True],
            False,
            ValueError,
        ),
        (
            make_index[True, False],
            False,
            ValueError,
        ),
        (
            make_index[True, False, True],
            False,
            ValueError,
        ),
        (
            make_index[None],
            False,
            ValueError,
        ),
        (
            make_index[None, None],
            False,
            ValueError,
        ),
        (
            make_index[None, None, None],
            False,
            ValueError,
        ),
        # Range selection
        (
            make_index[:, :],
            False,
            (0, 0, 10, 20),
        ),
        (
            make_index[5:, :],
            False,
            (0, 5, 10, 15),
        ),
        (
            make_index[:5, :],
            False,
            (0, 0, 10, 5),
        ),
        (
            make_index[:, 5:],
            False,
            (5, 0, 5, 20),
        ),
        (
            make_index[:, :5],
            False,
            (0, 0, 5, 20),
        ),
        # Range selection, but negative
        (
            make_index[-2:, :],
            False,
            (0, 18, 10, 2),
        ),
        (
            make_index[:-2, :],
            False,
            (0, 0, 10, 18),
        ),
        (
            make_index[:, -2:],
            False,
            (8, 0, 2, 20),
        ),
        (
            make_index[:, :-2],
            False,
            (0, 0, 8, 20),
        ),
        # Out of bounds
        (
            make_index[-50, :],
            False,
            IndexError,
        ),
        (
            make_index[:, -50],
            False,
            IndexError,
        ),
        # Ranged out of bounds
        (
            make_index[-60:, :],
            False,
            IndexError,
        ),
        (
            make_index[:-60, :],
            False,
            IndexError,
        ),
        (
            make_index[:, -60:],
            False,
            IndexError,
        ),
        (
            make_index[:, :-60],
            False,
            IndexError,
        ),
        # Incorrect types inside of slices
        (
            make_index[2.0:, :],
            False,
            ValueError,
        ),
        (
            make_index[:True, :],
            False,
            ValueError,
        ),
        (
            make_index[:, "Heyho":],
            False,
            ValueError,
        ),
        (
            make_index[:, :list],
            False,
            ValueError,
        ),
        # Indexing into the header
        (
            make_index["header", :],
            False,
            (0, "header", 10, 1),
        ),
        (
            make_index["header", 3:5],
            False,
            (3, "header", 2, 1),
        ),
        (
            make_index["header", -2:],
            False,
            (8, "header", 2, 1),
        ),
        (
            make_index["header", :-2],
            False,
            (0, "header", 8, 1),
        ),
        # Indexing into the header, but in the wrong axis
        (
            make_index[0, "header"],
            False,
            ValueError,
        ),
    ],
)
def test_indices(
    index: t.Any,
    enable_column_names: bool,
    result_should: tuple[
        int,
        int | t.Literal["header"],
        int,
        int,
    ]
    | t.Type[Exception],
) -> None:
    if enable_column_names:
        column_names = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"]
    else:
        column_names = None

    # Should this work, or fail?
    if isinstance(result_should, tuple):
        # Let Rio bring the selection into the format (left, top, width, height)
        result_is = rio.components.table._indices_to_rectangle(
            index=index,
            column_names=column_names,
            data_width=10,
            data_height=20,
        )

        # Does reality match expectations?
        assert result_is == result_should

    else:
        with pytest.raises(result_should):
            rio.components.table._indices_to_rectangle(
                index,
                column_names=column_names,
                data_width=10,
                data_height=20,
            )



from ...components.fundamental_component import FundamentalComponent

__all__ = [
    "DevToolsConnector",
]


class DevToolsConnector(FundamentalComponent):
    pass


DevToolsConnector._unique_id_ = "DevToolsConnector-builtin"



import asyncio
import typing as t

import rio.testing
from rio.components import Component


class ChildMounter(rio.Component):
    child: rio.Component
    child_mounted: bool = False

    def toggle(self) -> None:
        self.child_mounted = not self.child_mounted

    def build(self) -> rio.Component:
        if self.child_mounted:
            return self.child
        else:
            return rio.Spacer()


class EventCounter(rio.Component):
    child: rio.Component

    # Rio mustn't interpret these as state, otherwise it will cause unexpected
    # rebuilds.
    if t.TYPE_CHECKING:
        mount_count: int = 0
        unmount_count: int = 0

    def __post_init__(self):
        self.mount_count = 0
        self.unmount_count = 0

    @rio.event.on_mount
    def _on_mount(self):
        self.mount_count += 1

    @rio.event.on_unmount
    def _on_unmount(self):
        self.unmount_count += 1

    def build(self) -> Component:
        return self.child


async def test_mounted():
    class NestedComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Text("hi")

    def build():
        return ChildMounter(EventCounter(NestedComponent()))

    async with rio.testing.TestClient(build) as test_client:
        mounter = test_client.get_component(ChildMounter)
        event_counter = t.cast(EventCounter, mounter.child)
        assert event_counter.mount_count == 0
        assert event_counter.unmount_count == 0

        mounter.toggle()
        await test_client.refresh()
        assert event_counter.mount_count == 1
        assert event_counter.unmount_count == 0

        # Make sure the newly mounted components were sent to the client
        nested_component = test_client.get_component(NestedComponent)
        text_component = test_client.get_component(rio.Text)
        assert test_client._last_updated_components == {
            mounter,
            event_counter,
            nested_component,
            text_component,
        }

        mounter.toggle()
        await test_client.refresh()
        assert event_counter.unmount_count == 1


async def test_double_mount():
    def build():
        return ChildMounter(EventCounter(rio.Text("hello!")))

    async with rio.testing.TestClient(build) as test_client:
        mounter = test_client.get_component(ChildMounter)
        event_counter = t.cast(EventCounter, mounter.child)

        for _ in range(4):
            mounter.toggle()
            await test_client.refresh()

            if mounter.child_mounted:
                assert event_counter in test_client._last_updated_components

        assert event_counter.mount_count == 2
        assert event_counter.unmount_count == 2


async def test_refresh_after_synchronous_mount_handler():
    class DemoComponent(rio.Component):
        mounted: bool = False

        @rio.event.on_mount
        def on_mount(self):
            self.mounted = True

        def build(self) -> rio.Component:
            return rio.Switch(self.mounted)

    async with rio.testing.TestClient(DemoComponent) as test_client:
        demo_component = test_client.get_component(DemoComponent)
        switch = test_client.get_component(rio.Switch)

        assert demo_component.mounted

        last_component_state_changes = test_client._last_component_state_changes
        assert switch in last_component_state_changes
        assert last_component_state_changes[switch].get("is_on") is True


async def test_periodic():
    ticks = 0

    class DemoComponent(rio.Component):
        @rio.event.periodic(0.05)
        def tick(self):
            nonlocal ticks
            ticks += 1

        def build(self) -> rio.Component:
            return rio.Spacer()

    async with rio.testing.TestClient(DemoComponent) as test_client:
        ticks_before = ticks
        await asyncio.sleep(0.1)
        ticks_after = ticks
        assert ticks_after > ticks_before

        await test_client._simulate_interrupted_connection()

        ticks_before = ticks
        await asyncio.sleep(0.1)
        ticks_after = ticks
        assert ticks_after == ticks_before

        await test_client._simulate_reconnect()

        ticks_before = ticks
        await asyncio.sleep(0.1)
        ticks_after = ticks
        assert ticks_after > ticks_before



"""
Matches URL patterns to URLs, verifying that they match what they should.
"""

import typing as t
import warnings

import pytest

import rio
import rio.app_server
from rio.url_pattern import UrlPattern


@pytest.mark.parametrize(
    "pattern, exception",
    [
        # Leading slash
        (
            "/foo",
            ValueError,
        ),
        # Trailing slash
        (
            "foo/",
            ValueError,
        ),
        # Just slashes
        (
            "/",
            ValueError,
        ),
        (
            "//",
            ValueError,
        ),
        # Multiple consecutive slashes
        (
            "foo//bar",
            ValueError,
        ),
        (
            "foo///bar",
            ValueError,
        ),
        # Unterminated path parameter
        (
            "foo/{param",
            ValueError,
        ),
        (
            "foo/{param:path",
            ValueError,
        ),
        (
            "foo/param}",
            ValueError,
        ),
        (
            "foo/param:path}",
            ValueError,
        ),
        # Invalid path parameter names
        (
            "{???}",
            ValueError,
        ),
        (
            "{param}/{param}",
            ValueError,
        ),
    ],
)
def test_invalid_url_pattern(
    pattern: str,
    exception: t.Type[Exception],
) -> None:
    """
    Tests that invalid patterns raise the correct exceptions.
    """
    with pytest.raises(exception):
        UrlPattern(pattern)


@pytest.mark.parametrize(
    "url, pattern_str, values_should, remaining_should",
    [
        # Empty match (as used in the home page)
        (
            "",
            "",
            {},
            "",
        ),
        # Single segment
        (
            "foo",
            "foo",
            {},
            "",
        ),
        (
            "foo/bar",
            "foo",
            {},
            "bar",
        ),
        # Multiple segments
        (
            "foo/bar/baz",
            "foo/bar",
            {},
            "baz",
        ),
        # Single path parameter
        (
            "foo",
            "{param}",
            {"param": "foo"},
            "",
        ),
        (
            "foo/bar",
            "{param}",
            {"param": "foo"},
            "bar",
        ),
        (
            "foo/bar",
            "foo/{param}",
            {"param": "bar"},
            "",
        ),
        (
            "foo/bar/baz",
            "foo/{param}",
            {"param": "bar"},
            "baz",
        ),
        (
            "foo/bar/baz/qux",
            "foo/{param}/baz",
            {"param": "bar"},
            "qux",
        ),
        # Multiple path parameters
        (
            "foo/bar/baz",
            "foo/{param1}/{param2}",
            {"param1": "bar", "param2": "baz"},
            "",
        ),
        (
            "foo/bar/baz/qux/quux",
            "foo/{param1}/baz/{param2}",
            {"param1": "bar", "param2": "qux"},
            "quux",
        ),
        # "Greedy" path parameter
        (
            "foo/bar/baz",
            "{param:path}",
            {"param": "foo/bar/baz"},
            "",
        ),
        (
            "foo/bar/baz",
            "foo/{param:path}",
            {"param": "bar/baz"},
            "",
        ),
        (
            "foo/bar/baz/qux/quux",
            "foo/{param1}/{param2:path}",
            {"param1": "bar", "param2": "baz/qux/quux"},
            "",
        ),
        (
            "foo/bar/baz/qux/quux",
            "{param:path}/qux",
            {"param": "foo/bar/baz"},
            "quux",
        ),
    ],
)
def test_url_pattern_should_match(
    url: str,
    pattern_str: str,
    values_should: dict[str, str],
    remaining_should: str,
) -> None:
    """
    Matches patterns to URLs and verifies that they match what they should. All
    provided patterns must match their URLs.
    """
    # Parse the pattern
    pattern_object = UrlPattern(pattern_str)

    # Match the URL
    did_match, values_are, remaining_are = pattern_object.match(url)

    # Verify that the URL matched
    assert did_match

    # Verify that the values are correct
    assert values_are == values_should

    # Verify that the remaining part of the URL is correct
    assert remaining_are == remaining_should


@pytest.mark.parametrize(
    "url, pattern_str",
    [
        (
            "foo",
            "",
        ),
        (
            "foo",
            "bar",
        ),
        (
            "foo",
            "foo/bar",
        ),
        (
            "users/foo",
            "user/{user_id}",
        ),
        (
            "",
            "{user_id}",
        ),
    ],
)
def test_url_pattern_should_not_match(
    url: str,
    pattern_str: str,
) -> None:
    """
    Matches patterns to URLs, verifying that they **don't** match.
    """
    # Parse the pattern
    pattern_object = UrlPattern(pattern_str)

    # Match the URL
    matched, _, _ = pattern_object.match(url)

    # Verify that the URL did not match
    assert not matched


def _build_empty() -> None:
    pass


def _build_and_verify_bool(param: rio.QueryParameter[bool] = True) -> None:
    assert isinstance(param, bool)


def _build_and_verify_int(param: rio.QueryParameter[int] = 123) -> None:
    assert isinstance(param, int) and not isinstance(param, (bool, float))


def _build_and_verify_float(param: rio.QueryParameter[float] = 123.4) -> None:
    assert isinstance(param, float) and not isinstance(param, (bool, int))


def _build_and_verify_str(
    param: rio.QueryParameter[str] = "default-value",
) -> None:
    assert isinstance(param, str)


def _build_and_verify_optional_bool(
    param: rio.QueryParameter[bool | None] = None,
) -> None:
    assert param is None or isinstance(param, bool)


def _build_and_verify_string_literal(
    param: rio.QueryParameter[t.Literal["foo", "bar"]] = "bar",
) -> None:
    assert param in ("foo", "bar")


def _build_and_verify_int_literal(
    param: rio.QueryParameter[t.Literal[1, 2, 3]] = 1,
) -> None:
    assert param in (1, 2, 3)


def _build_and_verify_float_literal(
    param: rio.QueryParameter[t.Literal[1.2, 3.4, 5.6]] = 1.2,  # type: ignore (are you kidding me, pyright?!)
) -> None:
    assert param in (1.2, 3.4, 5.6)


def _build_and_verify_bool_literal(
    param: rio.QueryParameter[t.Literal[False]] = False,
) -> None:
    assert param is False


@pytest.mark.parametrize(
    "url_str, pattern, build_function, kwargs_should",
    [
        (
            "foo/bar",
            "foo/bar",
            _build_empty,
            {},
        ),
        # Single path parameters of all types
        (
            "true",
            "{param}",
            _build_and_verify_bool,
            {"param": True},
        ),
        (
            "false",
            "{param}",
            _build_and_verify_bool,
            {"param": False},
        ),
        (
            "-1",
            "{param}",
            _build_and_verify_int,
            {"param": -1},
        ),
        (
            "0",
            "{param}",
            _build_and_verify_int,
            {"param": 0},
        ),
        (
            "1",
            "{param}",
            _build_and_verify_int,
            {"param": 1},
        ),
        (
            "-1.0",
            "{param}",
            _build_and_verify_float,
            {"param": -1.0},
        ),
        (
            "0.0",
            "{param}",
            _build_and_verify_float,
            {"param": 0.0},
        ),
        (
            "1.0",
            "{param}",
            _build_and_verify_float,
            {"param": 1.0},
        ),
        (
            "foo",
            "{param}",
            _build_and_verify_str,
            {"param": "foo"},
        ),
        (
            "false",
            "{param}",
            _build_and_verify_optional_bool,
            {"param": False},
        ),
        (
            "foo",
            "{param}",
            _build_and_verify_string_literal,
            {"param": "foo"},
        ),
        (
            "bar",
            "{param}",
            _build_and_verify_string_literal,
            {"param": "bar"},
        ),
        (
            "3",
            "{param}",
            _build_and_verify_int_literal,
            {"param": 3},
        ),
        (
            "3.4",
            "{param}",
            _build_and_verify_float_literal,
            {"param": 3.4},
        ),
        (
            "no",
            "{param}",
            _build_and_verify_bool_literal,
            {"param": False},
        ),
        # Query parameters
        (
            "foo?param=baz",
            "foo",
            _build_and_verify_str,
            {"param": "baz"},
        ),
        # Default fallback: Missing parameter
        #
        # The current implementation simply doesn't pass these values to the
        # build function, relying on Python to impute the default. Hence the
        # empty expected dictionaries.
        (
            "foo",
            "foo",
            _build_and_verify_bool,
            {},
        ),
        (
            "foo",
            "foo",
            _build_and_verify_int,
            {},
        ),
        (
            "foo",
            "foo",
            _build_and_verify_float,
            {},
        ),
        (
            "foo",
            "foo",
            _build_and_verify_str,
            {},
        ),
        (
            "foo",
            "foo",
            _build_and_verify_string_literal,
            {},
        ),
        # Default fallback: Invalid parameter value
        (
            "foo?param=bar",
            "foo",
            _build_and_verify_bool,
            {},
        ),
        (
            "foo?param=bar",
            "foo",
            _build_and_verify_int,
            {},
        ),
        (
            "foo?param=bar",
            "foo",
            _build_and_verify_float,
            {},
        ),
        (
            "foo?param=hello",
            "foo",
            _build_and_verify_string_literal,
            {},
        ),
        (
            "foo?param=yes",
            "foo",
            _build_and_verify_bool_literal,
            {},
        ),
        # TODO: Multiple parameters
    ],
)
def test_parse_url_kwargs(
    url_str: str,
    pattern: str,
    build_function: t.Callable[..., t.Any],
    kwargs_should: dict[str, object],
) -> None:
    """
    Matches URLs against patterns, ensuring the parameters are correctly
    extracted and parsed.
    """
    # Create a component page object
    page = rio.ComponentPage(
        name="Test Page",
        url_segment=pattern,
        build=build_function,
    )

    # Parse the provided URL
    url = rio.URL(url_str)
    did_match, raw_path_arguments, remaining_url = page._url_pattern.match(
        url.path,
    )

    assert did_match
    assert remaining_url == ""

    # Get the kwargs for the build function
    kwargs_are = {
        name: page._url_parameter_parsers[name].parse(value)
        for name, value in raw_path_arguments.items()
    }
    kwargs_are.update(page._parse_query_parameters(url.query))

    # Verify that the kwargs are correct
    assert kwargs_are == kwargs_should


def test_layout_parameters_arent_url_parameters() -> None:
    # When a Component is used as the `build` function, layout properties like
    # `grow_x` must not be controllable via the URL.
    class MyComponent(rio.Component):
        foo: rio.QueryParameter[int] = 0

        def build(self) -> rio.Component:
            return rio.Text(str(self.foo))

    page = rio.ComponentPage(
        name="Test Page",
        url_segment="foobar",
        build=MyComponent,
    )

    kwargs = page._parse_query_parameters(
        {
            "foo": "7",
            "key": "0",
            "grow_x": "true",
            "grow_y": "1",
            "min_width": "5",
            "min_height": "5",
            "margin": "5",
            "_id_": "5",
        },
    )

    assert kwargs == {"foo": 7}


def test_annotations_dont_have_to_be_resolvable():
    # The type annotation is invalid, but it's quite clearly not supposed to be
    # a query parameter, so it should be silently ignored.
    def build(foo: "oops" = 0):  # type: ignore
        return rio.Text(str(foo))

    with warnings.catch_warnings(record=True) as warnings_list:
        rio.ComponentPage(
            name="Test Page",
            url_segment="foobar",
            build=build,
        )

    assert not warnings_list


def test_unresolvable_annotation_warning():
    # The type annotation is invalid, but it's likely meant to be a query
    # parameter, so rio should emit a warning.
    def build(foo: "QueryParameter[int]" = 0):  # type: ignore
        return rio.Text(str(foo))

    with warnings.catch_warnings(record=True) as warnings_list:
        rio.ComponentPage(
            name="Test Page",
            url_segment="foobar",
            build=build,
        )

    assert len(warnings_list) == 1


def test_valid_urls_list():
    def build_search(query: rio.QueryParameter[str] = ""):
        return rio.Text(query)

    def build_user_profile_page(user_id: int):
        return rio.Text(str(user_id))

    def build_switch_page(switch: bool):
        return rio.Switch(switch)

    def build_docs(
        version: t.Literal[1, 2, 3],
        section: t.Literal["tutorial", "api"],
    ):
        return rio.Text(f"{section} v{version}")

    def build_docs_subpage(param: t.Literal["foo", "bar"]):
        return rio.Text(param)

    app = rio.App(
        name="Test App",
        pages=[
            # Static URL
            rio.ComponentPage(
                name="Home",
                url_segment="",
                build=rio.Spacer,
            ),
            # Static URL + query parameter
            rio.ComponentPage(
                name="Search",
                url_segment="search",
                build=build_search,
            ),
            # Dynamic URL with `int` parameter
            rio.ComponentPage(
                name="Users",
                url_segment="user/{user_id}",
                build=build_user_profile_page,
            ),
            # Dynamic URL with `bool` parameter
            rio.ComponentPage(
                name="Switch",
                url_segment="switch/{switch}",
                build=build_switch_page,
            ),
            # Dynamic URL with literal parameters *and* subpages
            rio.ComponentPage(
                name="Docs",
                url_segment="docs/{version}/{section}",
                build=build_docs,
                children=[
                    rio.ComponentPage(
                        name="Docs Subpage",
                        url_segment="{param}",
                        build=build_docs_subpage,
                    ),
                ],
            ),
        ],
    )

    urls = set[str]()
    for url in app._iter_page_urls(include_redirects=False):
        assert url not in urls, f"Duplicate URL: {url}"
        urls.add(url)

    assert urls == {
        "",
        "search",
        # No `user/` url because its parameter is an int, and there are infinite
        # ints
        "switch/true",
        "switch/false",
        # No bare `docs/x/y` because there is no subpage with `url_segment=""`
        "docs/1/tutorial/foo",
        "docs/1/tutorial/bar",
        "docs/1/api/foo",
        "docs/1/api/bar",
        "docs/2/tutorial/foo",
        "docs/2/tutorial/bar",
        "docs/2/api/foo",
        "docs/2/api/bar",
        "docs/3/tutorial/foo",
        "docs/3/tutorial/bar",
        "docs/3/api/foo",
        "docs/3/api/bar",
    }



import os
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

import build as build_frontend
import revel

import rio.cli.project_setup
import rio.snippets
from rio.version import Version


def main() -> None:
    # Sanity checks
    revel.print_chapter("Running sanity checks")

    ensure_branch("main")
    ensure_no_uncommitted_changes()
    ensure_up_to_date_with_remote()

    build_frontend.build(mode="release")
    ensure_tests_pass()

    revel.success("Everything is in order.")

    # Update the templates repository
    revel.print_chapter("Updating templates repository")
    publish_current_templates()

    # Publish
    revel.print_chapter("Building & Publishing release")
    make_new_release()


def ensure_branch(branch_name: str) -> None:
    process = subprocess.run(
        ["git", "branch", "--show-current"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip("\n") != branch_name:
        revel.fatal(f'You must checkout the "{branch_name}" branch')


def ensure_no_uncommitted_changes() -> None:
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("There are uncommitted changes")


def ensure_up_to_date_with_remote() -> None:
    subprocess.run(["git", "remote", "update"], check=True)
    process = subprocess.run(
        ["git", "status", "--porcelain", "--untracked-files=no"],
        check=True,
        capture_output=True,
        text=True,
    )
    if process.stdout.strip():
        revel.fatal("Local branch is not up-to-date with remote")


def ensure_tests_pass() -> None:
    process = subprocess.run(
        ["uv", "run", "--", "pytest", "tests", "-x", "--disable-warnings"]
    )

    if process.returncode == 0:
        revel.print("Everything is in order. Publishing...")
    else:
        if not revel.select_yes_no(
            "The test suite does not pass. Do you want to publish anyway?"
        ):
            sys.exit(1)

        revel.print("Publishing...")


def has_git_changes() -> bool:
    result = subprocess.run(
        ["git", "status", "--porcelain"],
        capture_output=True,
        text=True,
        check=True,
    )
    return bool(result.stdout.strip())


def publish_current_templates() -> None:
    """
    All of Rio's built-in templates are available in source code form in a
    public repository. This allows users to browse the code. This function
    ensures that the latest versions of all templates are available in the
    repository.

    It's unclear whether this function should be in the `rio` project itself, or
    the website. While not strongly linked to the website, having the repository
    update here ensures that the links on the Rio website are always correct,
    i.e. all templates that the site links to are also definitely in that
    repository.
    """
    # Check out the repository
    template_repo_dir = Path(tempfile.mkdtemp())

    subprocess.run(
        [
            "git",
            "clone",
            "git@github.com:rio-labs/rio-templates.git",
            str(template_repo_dir),
        ],
        check=True,
    )

    # Wipe the current templates
    keep_files = (
        ".git",
        "LICENSE.txt",
        "README.md",
    )

    for file_path in template_repo_dir.iterdir():
        if file_path.name not in keep_files:
            assert file_path.is_dir(), file_path
            shutil.rmtree(file_path)

    # Save the current working directory
    original_working_directory = os.getcwd()

    # Re-instantiate all templates
    #
    # This prints profusely, so tell it to be quiet
    original_stdout = sys.stdout
    sys.stdout = open(os.devnull, "w")

    os.chdir(template_repo_dir)

    for template in rio.snippets.get_project_templates(include_empty=False):
        rio.cli.project_setup.create_project(
            raw_name=template.name,
            type="website",
            template_name=template.name,
            target_parent_directory=template_repo_dir,
        )

    # Commit and push, if there are changes
    subprocess.run(["git", "add", "."], check=True)

    if has_git_changes():
        subprocess.run(["git", "commit", "-m", "Update templates"], check=True)
        subprocess.run(["git", "push"], check=True)

    # Restore previous state
    sys.stdout = original_stdout
    os.chdir(original_working_directory)


def make_new_release() -> None:
    # Bump version
    bump_version()
    subprocess.run(["git", "push"], check=True)

    # Publish
    subprocess.run(["uv", "build"], check=True)
    subprocess.run(["uv", "publish"], check=True)

    # Create a tag
    version = str(get_version())
    subprocess.run(["git", "tag", version], check=True)
    subprocess.run(["git", "push", "origin", "tag", version], check=True)


def bump_version() -> None:
    version = get_version()

    options = dict[str, str]()

    if version.rc is None:
        v = version.bump_patch()
        options[f"Patch:    {v}"] = str(v)

        v = v.bump_rc()
        options[f"Patch RC: {v}"] = str(v)

        v = version.bump_minor()
        options[f"Minor:    {v}"] = str(v)

        v = v.bump_rc()
        options[f"Minor RC: {v}"] = str(v)

        v = version.bump_major()
        options[f"Major:    {v}"] = str(v)

        v = v.bump_rc()
        options[f"Major RC: {v}"] = str(v)
    else:
        v = version.bump_rc()
        options[f"RC:    {v}"] = str(v)

        v = version.drop_rc()
        options[f"No RC: {v}"] = str(v)

    new_version = revel.select(options, prompt="Select new version")

    subprocess.run(
        [sys.executable, "-m", "hatch", "version", new_version],
        check=True,
    )
    subprocess.run(
        ["git", "commit", "--all", "-m", f"bump version to {new_version}"],
        check=True,
    )


def get_version() -> Version:
    version_string = subprocess.run(
        [sys.executable, "-m", "hatch", "version"],
        check=True,
        capture_output=True,
        text=True,
    ).stdout.strip()

    return Version.parse(version_string)


if __name__ == "__main__":
    main()



import rio

# <additional-imports>
from . import data_models, theme

# </additional-imports>


# <additional-code>
def on_session_start(sess: rio.Session) -> None:
    # Determine which layout to use
    if sess.window_width < 60:
        layout = data_models.PageLayout(
            device="mobile",
        )
    else:
        layout = data_models.PageLayout(
            device="desktop",
        )

    # Attach the layout to the session
    sess.attach(layout)


# </additional-code>


# Make sure ruff doesn't remove unused imports
# Create the Rio app
app = rio.App(
    name="crypto_dashboard",
    theme=theme.THEME,
)



import asyncio

import fastapi
import typing_extensions as te

from . import abstract_transport

__all__ = ["FastapiWebsocketTransport"]


class FastapiWebsocketTransport(abstract_transport.AbstractTransport):
    def __init__(self, websocket: fastapi.WebSocket):
        super().__init__()

        self._websocket = websocket
        self._closed_intentionally = False

    @te.override
    async def send(self, msg: str) -> None:
        try:
            await self._websocket.send_text(msg)
        except RuntimeError:
            pass  # Socket is already closed

    @te.override
    async def receive(self) -> str:
        try:
            return await self._websocket.receive_text()
        except RuntimeError:
            pass  # Socket is already closed
        except fastapi.WebSocketDisconnect as err:
            self._closed_intentionally = err.code == 1001

        if self._closed_intentionally:
            raise abstract_transport.TransportClosedIntentionally
        else:
            raise abstract_transport.TransportInterrupted

    def close(self) -> None:
        self._closed_intentionally = True
        asyncio.create_task(self._close_websocket())

    async def _close_websocket(self):
        try:
            await self._websocket.close()
        except RuntimeError:
            pass  # websocket already closed

        self.closed.set()



from . import data_models


class Persistence:
    """
    A class to handle operations for the MenuItems.

    You can adapt this class to your needs by adding more methods to interact
    with the "database" or support different databases like SQLight or MongoDB.

    ## Attributes

    `menu_items`: The list of MenuItem objects.
    """

    def __init__(self) -> None:
        """
        Initializes the Persistence object with a list of MenuItem objects.
        """
        self.menu_items: list[data_models.MenuItem] = (
            self._populate_menu_items()
        )

    def _populate_menu_items(self) -> list[data_models.MenuItem]:
        """
        Populates the initial list of MenuItem objects.
        """
        return [
            data_models.MenuItem(
                name="Hamburger",
                description="A classic hamburger with lettuce, tomato, and onions",
                price=4.99,
                category="Burgers",
            ),
            data_models.MenuItem(
                name="Cheeseburger",
                description="A classic cheeseburger with lettuce, tomato, onions and cheese",
                price=5.99,
                category="Burgers",
            ),
            data_models.MenuItem(
                name="Fries",
                description="A side of crispy fries",
                price=2.99,
                category="Sides",
            ),
            data_models.MenuItem(
                name="Soda",
                description="A refreshing soda",
                price=1.99,
                category="Drinks",
            ),
            data_models.MenuItem(
                name="Salad",
                description="A fresh salad",
                price=4.99,
                category="Salads",
            ),
            data_models.MenuItem(
                name="Ice Cream",
                description="A sweet treat",
                price=3.99,
                category="Desserts",
            ),
        ]



import rio.testing


async def test_refresh_with_nothing_to_do() -> None:
    def build() -> rio.Component:
        return rio.Text("Hello")

    async with rio.testing.TestClient(build) as test_client:
        test_client._outgoing_messages.clear()
        await test_client.refresh()

        assert not test_client._dirty_components
        assert not test_client._last_updated_components


async def test_refresh_with_clean_root_component() -> None:
    def build() -> rio.Component:
        text_component = rio.Text("Hello")
        return rio.Container(text_component)

    async with rio.testing.TestClient(build) as test_client:
        text_component = test_client.get_component(rio.Text)

        text_component.text = "World"
        await test_client.refresh()

        assert test_client._last_updated_components == {text_component}


async def test_rebuild_component_with_dead_parent() -> None:
    class ChildUnmounter(rio.Component):
        child: rio.Component
        child_is_mounted: bool = True

        def build(self) -> rio.Component:
            if self.child_is_mounted:
                return self.child
            else:
                return rio.Spacer()

    class ComponentWithState(rio.Component):
        state: str

        def build(self) -> rio.Component:
            return rio.Text(self.state)

    def build() -> rio.Component:
        return ChildUnmounter(ComponentWithState("Hello"))

    async with rio.testing.TestClient(
        build, use_ordered_dirty_set=True
    ) as test_client:
        # Change the component's state, but also remove it from the component
        # tree
        unmounter = test_client.get_component(ChildUnmounter)
        component = test_client.get_component(ComponentWithState)

        component.state = "Hi"
        unmounter.child_is_mounted = False

        await test_client.refresh()

        # Make sure no data for dead components was sent to JS
        assert unmounter in test_client._last_updated_components
        assert component not in test_client._last_updated_components


async def test_unmount_and_remount() -> None:
    class DemoComponent(rio.Component):
        content: rio.Component
        show_child: bool

        def build(self) -> rio.Component:
            children = [self.content] if self.show_child else []
            return rio.Row(*children)

    def build() -> rio.Component:
        return DemoComponent(
            rio.Text("hi"),
            show_child=True,
        )

    async with rio.testing.TestClient(build) as test_client:
        root_component = test_client.get_component(DemoComponent)
        child_component = root_component.content
        row_component = test_client.get_component(rio.Row)

        root_component.show_child = False
        await test_client.refresh()
        assert test_client._last_updated_components == {
            root_component,
            row_component,
        }

        root_component.show_child = True
        await test_client.refresh()
        assert test_client._last_updated_components == {
            root_component,
            row_component,
            child_component,
        }



import rio

# <additional-imports>
from . import persistence

# </additional-imports>


# <additional-code>
def on_app_start(app: rio.App) -> None:
    """
    A function that runs when the app is started.
    """
    # Create a persistence instance. This class hides the gritty details of
    # database interaction from the app.
    pers = persistence.Persistence()

    # Now attach it to the session. This way, the persistence instance is
    # available to all components using `self.session[persistence.Persistence]`
    app.default_attachments.append(pers)


# </additional-code>


# <additional-demo-code>
def on_demo_session_start(sess: rio.Session) -> None:
    """
    A function that runs when a new session is started.
    """
    # Create a persistence instance.
    pers = persistence.Persistence()

    # Attach the persistence instance to the session.
    sess.attach(pers)


# </additional-demo-code>



from __future__ import annotations

import dataclasses
import typing as t

from uniserde import JsonDoc

import rio

from .. import deprecations, text_style, utils
from .fundamental_component import FundamentalComponent

__all__ = [
    "Text",
]


@t.final
class Text(FundamentalComponent):
    """
    Displays unformatted text.

    `Text` displays text without any formatting, making it one of the most
    commonly used components in Rio.

    While the text itself is unformatted, you can still control the style of
    the text using the `style` attribute. This allows you to change the font
    size, color, and more.


    ## Attributes

    `text`: The text to display.

    `selectable`: Whether the text can be selected by the user.

    `style`: The style of the text. This can either be a `TextStyle` instance,
        or one of the built-in styles: `"heading1"`, `"heading2"`, `"heading3"`,
        `"text"` or `"dim"`.

    `justify`: Controls the alignment of the text within each individual line.

    `wrap`: Deprecated in favor of `overflow`.

    `overflow`: What to do with long text. If this is `"nowrap"`, the text will
        be as wide as it needs to be to fit without wrapping. If `"wrap"`, the
        text will wrap to the next line when running out of horizontal space.
        Finally, if `"ellipsize"`, the text will be truncated when there isn't
        enough space and an ellipsis (`...`) will be added.

        `font`: The `Font` to use for the text. When set to `None`, the default font
        for the current context (heading or regular text, etc) will be used.

    `font`: The name of the font to use. Overrides the `font` from the `style`.

    `fill`: The fill (color, gradient, etc.) for the text. Overrides the `fill`
        from the `style`.

    `font_size`: The font size. Overrides the `font_size` from the `style`.

    `italic`: Whether the text is *italic* or not. Overrides the `italic` from
        the `style`.

    `font_weight`: Whether the text is normal or **bold**. Overrides the
        `font_weight` from the `style`.

    `underlined`: Whether the text is underlined or not. Overrides the
        `underlined` from the `style`.

    `strikethrough`: Whether the text should have a line through it. Overrides
        the `strikethrough` from the `style`.

    `all_caps`: Whether the text is transformed to ALL CAPS or not. Overrides
        the `all_caps` from the `style`.


    ## Examples

    Here's a minimal example for displaying text. Just pass it a string:

    ```python
    rio.Text("Hello, world!")
    ```

    To change the style of the text, you can pass a `TextStyle` instance:

    ```python
    rio.Text(
        "Hello, world!",
        style=rio.TextStyle(
            font_size=3,
            # Text Style has a lot of optional parameters. Have a look at its
            # docs for all details!
            #
            # https://rio.dev/docs/api/textstyle
        ),
    )
    ```
    """

    text: str
    _: dataclasses.KW_ONLY
    selectable: bool = True
    style: (
        t.Literal["heading1", "heading2", "heading3", "text", "dim"]
        | rio.TextStyle
    ) = "text"
    justify: t.Literal["left", "right", "center", "justify"] = "left"
    wrap: bool | t.Literal["ellipsize"] = False
    overflow: t.Literal["nowrap", "wrap", "ellipsize"] = "nowrap"

    font: rio.Font | None = None
    fill: text_style._TextFill | None | utils.NotGiven = utils.NOT_GIVEN
    font_size: float | None = None
    italic: bool | None = None
    font_weight: t.Literal["normal", "bold"] | None = None
    underlined: bool | None = None
    strikethrough: bool | None = None
    all_caps: bool | None = None

    def _custom_serialize_(self) -> JsonDoc:
        # Serialization doesn't handle unions. Hence the custom serialization
        # here
        if isinstance(self.style, str):
            style = self.style
        else:
            style = self.style._serialize(self.session)

        # The old `wrap` attribute has been replaced with `overflow`. Remap the
        # value.
        if self.wrap is not False:
            deprecations.warn_parameter_renamed(
                since="0.10",
                old_name="wrap",
                new_name="overflow",
                function="rio.Text",
            )

        if self.wrap is True:
            overflow = "wrap"
        elif self.wrap == "ellipsize":
            overflow = "ellipsize"
        else:
            overflow = self.overflow

        if isinstance(self.fill, utils.NotGiven):
            fill = "not given"
        else:
            fill = self.session._serialize_fill(self.fill)

        # Build the result
        return {
            "style": style,
            "overflow": overflow,
            "fill": fill,
        }

    def __repr__(self) -> str:
        if len(self.text) > 40:
            text = self.text[:40] + "..."
        else:
            text = self.text

        return f"<{type(self).__name__} id:{self._id_} text:{text!r}>"


Text._unique_id_ = "Text-builtin"



import pytest

import rio
from tests.utils.layouting import verify_layout


@pytest.mark.parametrize(
    "justify",
    [
        "left",
        "right",
        "center",
        "justified",
        "grow",
    ],
)
async def test_flow_container_layout(justify: str) -> None:
    await verify_layout(
        lambda: rio.FlowContainer(
            rio.Text("foo", min_width=5),
            rio.Text("bar", min_width=10),
            rio.Text("qux", min_width=4),
            column_spacing=3,
            row_spacing=2,
            justify=justify,  # type: ignore
            min_width=20,
            align_x=0,
        )
    )



import asyncio
import threading
import time
import typing as t

import rio

from . import run_models


class WebViewWorker:
    def __init__(
        self,
        *,
        push_event: t.Callable[[run_models.Event], None],
        debug_mode: bool,
        url: str,
    ) -> None:
        from ... import webview_shim

        self.push_event = push_event
        self.debug_mode = debug_mode
        self.url = url

        # If running, this is the webview window
        self.window: webview_shim.Window | None = None

    def run(self, initial_app: rio.App) -> None:
        """
        Starts the worker and blocks until the window is closed or
        `request_stop` is called. This function must be called from the main
        thread.
        """
        from ... import webview_shim

        assert self.window is None, "Already running"

        # Make sure this was called from the main thread.
        assert (
            threading.current_thread() is threading.main_thread()
        ), "Must be called from the main thread"

        # Fetch the icon
        icon_path = asyncio.run(initial_app._fetch_icon_as_png_path())

        # Create the window
        self.window = webview_shim.create_window(
            title=self._title_for_app(initial_app),
            url=self.url,
        )
        webview_shim.start_mainloop(icon=icon_path)

        # If we've reached this point, the window must have been closed.
        self.window = None

        # If the webview window is closed, tell the Arbiter about it.
        self.push_event(run_models.StopRequested())

    def request_stop(self) -> None:
        """
        Makes the worker stop soon after called. May be called multiple times,
        or even before startup has finished.
        """

        while self.window is not None:
            # Try to close the window
            try:
                self.window.destroy()

            # Closing can fail if the window hasn't entirely opened yet. In that
            # case try again in a bit.
            except KeyError:  # TODO: Why key error?
                time.sleep(0.05)  # TODO

            # Success. Make sure to forget the window.
            else:
                self.window = None

    def _title_for_app(self, app: rio.App) -> str:
        return app.name + (" (debug)" if self.debug_mode else "")

    def update_window_for_app(self, app: rio.App) -> None:
        if self.window is None:
            return

        self.window.set_title(self._title_for_app(app))



from __future__ import annotations

import typing as t

import imy.docstrings

import rio

from .. import utils
from .component import Component

__all__ = [
    "DialogContainer",
]


@t.final
@imy.docstrings.mark_as_private
class DialogContainer(Component):
    build_content: t.Callable[[], Component]
    owning_component_id: int
    is_modal: bool
    is_user_closable: bool
    on_close: rio.EventHandler[[]]
    style: t.Literal["default", "custom"] = "default"

    def build(self) -> Component:
        # Build the content
        content = utils.safe_build(self.build_content)

        # Apply the default dialog styling, if requested
        if self.style == "default":
            theme = self.session.theme

            content = rio.Rectangle(
                content=rio.Container(
                    rio.ThemeContextSwitcher(content, "neutral"),
                    margin=1.5,
                ),
                fill=theme.neutral_color,
                corner_radius=theme.corner_radius_medium,
                shadow_radius=1.8,
                shadow_offset_y=1.0,
                align_x=0.5,
                align_y=0.4,
            )

        # Done
        return content

    # Note that this is NOT `_custom_serialize`. Dialog containers are
    # high-level on the Python side, but sent to the client as though they were
    # fundamental. To prevent a whole bunch of custom code in the serializer,
    # this function handles the serialization of dialog containers.
    def serialize(self) -> dict[str, t.Any]:
        return {
            "owning_component_id": self.owning_component_id,
            "is_modal": self.is_modal,
            "is_user_closable": self.is_user_closable,
        }



import json

import pytest

import rio.arequests as arequests


def test_http_response_invalid_json() -> None:
    response = arequests.HttpResponse(
        status_code=200,
        headers={},
        content=b"invalid json",
    )

    with pytest.raises(json.JSONDecodeError, match="Expecting value"):
        response.json()


def test_http_response_invalid_utf8() -> None:
    response = arequests.HttpResponse(
        status_code=200,
        headers={},
        content=b"\xff",
    )

    with pytest.raises(json.JSONDecodeError, match="UTF-8"):
        response.json()


def test_request() -> None:
    response = arequests.request_sync(
        "get",
        "https://postman-echo.com/get",
        json={"foo": "bar"},
    )

    assert response.status_code == 200
    assert response.headers["content-type"] == "application/json; charset=utf-8"

    response_json = response.json()

    assert response_json["headers"]["user-agent"] == "rio.arequests/0.1"
    assert response_json["headers"]["host"] == "postman-echo.com"
    assert response_json["headers"]["content-type"] == "application/json"
    assert response_json["headers"]["content-length"] == "14"
    assert response_json["args"] == {"foo": "bar"}



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps

# </additional-imports>


# <component>
class RootComponent(rio.Component):
    """
    This page will be used as the root component for the app. This means, that
    it will always be visible, regardless of which page is currently active.

    This makes it the perfect place to put components that should be visible on
    all pages, such as a navbar or a footer.

    Additionally, the root page will contain a `rio.PageView`. Page views don't
    have any appearance on their own, but they are used to display the content
    of the currently active page. Thus, we'll always see the navbar and footer,
    with the content of the current page in between.
    """

    def build(self) -> rio.Component:
        return rio.Column(
            # The navbar contains a `rio.Overlay`, so it will always be on top
            # of all other components.
            comps.Navbar(),
            # Add some empty space so the navbar doesn't cover the content.
            rio.Spacer(min_height=10),
            # The page view will display the content of the current page.
            rio.PageView(
                # Make sure the page view takes up all available space.
                grow_y=True,
            ),
            # The footer is also common to all pages, so place it here.
            comps.Footer(),
        )


# </component>



from __future__ import annotations

# <additional-imports>
import rio

from .. import components as comps
from .. import data_models, theme

# </additional-imports>


# <component>
class HeroSection(rio.Component):
    """
    A responsive hero section that displays a call to action for building SaaS applications.

    The section includes:
    - A button linking to the Rio website.
    - A main title and subtitle.
    - A row of call-to-action buttons.
    - An image placeholder showcasing the hero visual.
    """

    def _desktop_build(self) -> rio.Component:
        """
        Builds the column layout optimized for desktop devices.
        """
        return rio.Column(
            # Add button linking to the Rio website
            comps.OutlinedColoredButton(
                text="Made with Rio",
                target_url="https://rio.dev",
                icon="material/open_in_new",
                align_x=0.5,
            ),
            # Add main title
            rio.Text(
                "Build your SaaS in seconds",
                justify="center",
                style=theme.BOLD_SECTION_TITEL_DESKTOP,
            ),
            # Add subtitle
            rio.Text(
                "This template demonstrates most of Rio's components. It's a great starting point to build your own SaaS.",
                justify="center",
                style=rio.TextStyle(
                    font_size=theme.ACTION_TEXT_HEIGHT,
                ),
            ),
            # Add call-to-action buttons
            rio.Row(
                rio.Button(
                    rio.Row(
                        rio.Text("Get started"),
                        rio.Icon("material/arrow_forward"),
                        align_x=0.5,
                        spacing=1,
                        margin_x=1,
                    ),
                    color=rio.Color.from_hex("ffffff"),
                ),
                comps.OutlinedNeutralButton(
                    text="Use this template", icon="thirdparty/github_logo"
                ),
                #
                spacing=1,
                align_x=0.5,
            ),
            # Add image placeholder
            comps.ImagePlaceholder(
                image_min_height=40,
                margin_top=5,
            ),
            margin_top=5,  # Top margin for the hero section
            spacing=2,  # Space between elements
            align_x=0.5,  # Center-align the column
            min_width=80,  # Fixed width for desktop layout
        )

    def _mobile_build(self) -> rio.Component:
        """
        Builds the column layout optimized for mobile devices.
        """
        return rio.Column(
            # Add button linking to the Rio website
            comps.OutlinedColoredButton(
                text="Made with Rio",
                target_url="https://rio.dev",
                icon="material/open_in_new",
                align_x=0.5,
            ),
            rio.Column(
                # Split main title so it fits on mobile screens with preferred
                # text wrapping
                rio.Text(
                    "Build your SaaS",
                    justify="center",
                    style=theme.BOLD_SMALLER_SECTION_TITLE_MOBILE,
                ),
                rio.Text(
                    "in seconds",
                    justify="center",
                    style=theme.BOLD_SMALLER_SECTION_TITLE_MOBILE,
                ),
            ),
            # Add subtitle
            rio.Text(
                "This template demonstrates most of Rio's components. It's a great starting point to build your own SaaS.",
                justify="center",
                overflow="wrap",
                style=rio.TextStyle(
                    font_size=theme.ACTION_TEXT_HEIGHT,
                ),
            ),
            # Add call-to-action buttons
            rio.Column(
                rio.Button(
                    rio.Row(
                        rio.Text("Get started"),
                        rio.Icon("material/arrow_forward"),
                        align_x=0.5,
                        spacing=1,
                    ),
                    color=rio.Color.from_hex("ffffff"),
                ),
                comps.OutlinedNeutralButton(
                    text="Use this template", icon="thirdparty/github_logo"
                ),
                align_x=0.5,
                spacing=1,
            ),
            # Add image placeholder
            comps.ImagePlaceholder(
                image_min_height=13,
                margin_top=3,
            ),
            spacing=2,
        )

    def build(self) -> rio.Component:
        """
        Build and return the page layout based on the user's device type.

        Determines whether the user is on a desktop or mobile device and builds
        the corresponding layout.
        """
        device = self.session[data_models.PageLayout].device

        if device == "desktop":
            # Build and return the desktop-optimized layout
            return self._desktop_build()

        else:
            # Build and return the mobile-optimized layout
            return self._mobile_build()


# </component>



from __future__ import annotations

import dataclasses
import typing as t
from datetime import date

import narwhals as nw
import narwhals.dtypes
import typing_extensions as te
from uniserde import JsonDoc

import rio

from .. import maybes
from .fundamental_component import FundamentalComponent

if t.TYPE_CHECKING:
    import numpy  # type: ignore
    import pandas  # type: ignore
    import polars  # type: ignore


__all__ = ["Table"]


# Datatype for any values which may be sent to the front-end. Any input types
# that don't jive with these have `str` applied to them to get them as strings.
NormalizedTableValue = int | float | str


@t.final
class Table(FundamentalComponent):
    """
    Display for tabular data.

    Tables are a way to display data in a grid, with rows and columns. They are
    well suited for displaying data that is naturally tabular, such as
    spreadsheets, databases, or CSV files.

    You can provide the data to tables in a variety of formats: `pandas`
    DataFrames, `polars` DataFrames, NumPy arrays, dictionaries, or lists of
    lists.

    If the data format provides header names (like DataFrames), they will be
    displayed at the top of the table.


    ## Attributes

    `data`: The data to display.

    `show_row_numbers`: Whether to show row numbers on the left side of the
        table.


    ## Examples

    A simple table with some data:

    ```python
    rio.Table(
        data={
            "Name": ["Alice", "Bob", "Charlie"],
            "Age": [25, 30, 35],
            "City": ["New York", "San Francisco", "Los Angeles"],
        }
    )
    ```

    Indexing into a table to apply styling works similar as numpy arrays.
    The following example makes the second and third row, and the second
    and third column bold:

    ```python
    class MyComponent(rio.Component):
        def build(self) -> rio.Component:
            table = rio.Table(
                data={
                    "Name": ["Alice", "Bob", "Charlie"],
                    "Age": [25, 30, 35],
                    "City": ["New York", "San Francisco", "Los Angeles"],
                }
            )

            # Apply some styling to the table, indexing rows and columns
            # works similar as numpy arrays.
            # Second and third row, second and third column are bold
            table[1:3, 1:3].style(font_weight="bold")

            return table
    ```


    ## Metadata

    `experimental`: True
    """

    data: (
        pandas.DataFrame
        | polars.DataFrame
        | t.Mapping[str, t.Iterable[t.Any]]
        | t.Iterable[t.Iterable[t.Any]]
        | numpy.ndarray
    )

    _: dataclasses.KW_ONLY

    show_row_numbers: bool = True

    # All headers, if present
    _headers: list[str] | None = dataclasses.field(default=None, init=False)

    # The data, as a list of columns ("column major"). This is set in
    # `__post_init__`.
    _columns: list[list[NormalizedTableValue]] = dataclasses.field(
        default_factory=list, init=False
    )

    # All styles applied to the table, in the same order they were added
    _styling: list[TableSelection] = dataclasses.field(
        default_factory=list, init=False
    )

    # These must be annotated, otherwise rio won't understand that tables have
    # child components and won't copy over the new values when two Tables are
    # reconciled.
    _children: list[rio.Component] = dataclasses.field(
        default_factory=list, init=False
    )

    _child_positions: list[tuple[int, int]] = dataclasses.field(
        default_factory=list, init=False
    )

    def __post_init__(self) -> None:
        # Bring the data into a standardized format
        self._headers, self._columns = _data_to_columnar(
            self.data,
            self.session._date_format_string,
        )

        # Help out the reconciler. This is needed to make sure new values aren't
        # silently dropped
        self._properties_set_by_creator_.update(
            [
                "_headers",
                "_columns",
                "_styling",
                "_children",
                "_child_positions",
            ]
        )

    def _custom_serialize_(self) -> JsonDoc:
        session = self.session

        return {
            "headers": self._headers,
            "columns": self._columns,
            "styling": [style._serialize(session) for style in self._styling],
            "children": [child._id_ for child in self._children],
            "childPositions": self._child_positions,
        }  # type: ignore

    def add(
        self,
        child: rio.Component,
        row: int,
        column: int | str,
    ) -> te.Self:
        """
        Add a child component to the table

        Adds a child to the table at the specified location. Note that unlike
        with grids, children in tables always take up exactly one cell. This is
        necessary to allow for sorting & filtering of tables (a planned
        feature).

        Note that this method returns the `Table` instance afterwards, allowing
        you to chain multiple `add` calls together for concise code.


        ## Parameters

        `child`: The child component to add to the table.

        `row`: The row in which to place the child.

        `column`: The column in which to place the child.


        ## Raises

        `ValueError`: If the column index is a string and the table doesn't have
            any headers.

        `KeyError`: If the column index is a string and the table doesn't have a
            column with that name.


        ## Metadata

        `public`: False
        """
        assert isinstance(child, rio.Component), child

        # Make sure the column is an integer, propagating any exceptions
        if isinstance(column, str):
            column = self._column_name_to_int(column)

        # Add the child
        self._children.append(child)
        self._child_positions.append(
            (row, column),
        )

        # Return self for chaining
        return self

    def _shape(self) -> tuple[int, int]:
        """
        Returns the (height, width) of the table data. This does not include
        headers! This is like numpy's shape but takes into account the many
        different types of data that can be passed to the data attribute.
        """
        try:
            return (len(self._columns[0]), len(self._columns))
        except IndexError:
            return (0, 0)

    def _column_name_to_int(self, column_name: str) -> int:
        if self._headers is None:
            raise ValueError(
                "Cannot index into this table using a column name, since it doesn't have any headers"
            )

        try:
            return self._headers.index(column_name)
        except ValueError:
            raise KeyError(
                f"This table doesn't have a column named {column_name!r}"
            )

    def __getitem__(
        self,
        index: (
            str
            | tuple[
                int | slice | str,
                int | slice | str,
            ]
        ),
    ) -> TableSelection:
        # Get the index as a tuple (top, left, height, width)
        data_height, data_width = self._shape()

        left, top, width, height = _indices_to_rectangle(
            index,
            self._headers,
            data_width,
            data_height,
        )

        # Verify the indices are within bounds
        right = left + width
        out_of_bounds = (left < 0 or left >= data_width) or (
            right < 0 or right > data_width
        )

        if top == "header":
            assert height == 1, height
        else:
            bottom = top + height
            out_of_bounds = (
                out_of_bounds
                or (top < 0 or top >= data_height)
                or (bottom < 0 or bottom > data_height)
            )

        if out_of_bounds:
            raise IndexError(
                f"The table index {index!r} is out of bounds for a table of size {data_height}x{data_width}"
            )

        # Construct the result
        result = TableSelection(
            _table=self,
            _left=left,
            _top=top,
            _width=width,
            _height=height,
            _font_weight=None,
        )

        # Keep track of it
        self._styling.append(result)

        # Done!
        return result


Table._unique_id_ = "Table-builtin"


@t.final
@dataclasses.dataclass
class TableSelection:
    _table: Table

    _left: int
    _top: int | t.Literal["header"]
    _width: int
    _height: int

    _font_color: rio.Color | None = None
    _background_color: rio.Color | None = None
    _italic: bool | None = None
    _font_weight: t.Literal["normal", "bold"] | None = None

    def style(
        self,
        *,
        font_color: rio.Color | None = None,
        background_color: rio.Color | None = None,
        italic: bool | None = None,
        font_weight: t.Literal["normal", "bold"] | None = None,
    ) -> Table:
        # Store the passed in values
        if font_color is not None:
            self._font_color = font_color

        if background_color is not None:
            self._background_color = background_color

        if italic is not None:
            self._italic = italic

        if font_weight is not None:
            self._font_weight = font_weight

        # Return the table to allow chaining
        return self._table

    def _serialize(self, sess: rio.Session) -> JsonDoc:
        # Some values are always present
        result: JsonDoc = {
            "left": self._left,
            "top": self._top,
            "width": self._width,
            "height": self._height,
        }

        # Include only set styles
        if self._font_color is not None:
            result["fontColor"] = self._font_color._serialize(sess)

        if self._background_color is not None:
            result["backgroundColor"] = self._background_color._serialize(sess)

        if self._italic is not None:
            result["italic"] = self._italic

        if self._font_weight is not None:
            result["fontWeight"] = self._font_weight

        # Done
        return result


def _index_to_start_and_extent(
    index: int | slice | str,
    size_in_axis: int,
    axis: t.Literal["x", "y"],
) -> tuple[
    int | t.Literal["header"],
    int,
]:
    """
    Given a one-axis `__getitem__` index, returns the start and extent of
    the slice as non-negative integers.

    This function is intentionally a separate function instead of a method
    to allow for easy unit-testing.
    """
    # Python considers booleans to be integers. Guard against them being used as
    # numbers here
    if isinstance(index, bool):
        raise ValueError(
            f"Table indices should be integers or slices, not {index!r}"
        )

    # Integers are easy
    if isinstance(index, int):
        start = index
        stop = index + 1

    # Slices need more work
    elif isinstance(index, slice):
        if index.start is None:
            start = 0
        elif isinstance(index.start, int) and not isinstance(index.start, bool):
            start = index.start
        else:
            raise ValueError(
                f"Table indices should be integers or slices, not {index!r}"
            )

        if index.stop is None:
            stop = size_in_axis
        elif isinstance(index.stop, int) and not isinstance(index.stop, bool):
            stop = index.stop
        else:
            raise ValueError(
                f"Table indices should be integers or slices, not {index!r}"
            )

    # In the y-axis, "header" is a valid index
    elif axis == "y" and index == "header":
        return "header", 1

    # Anything else is invalid
    else:
        raise ValueError(
            f"Table indices should be integers or slices, not {index!r}"
        )

    # Negative numbers count backwards from the end
    if start < 0:
        if start + size_in_axis < 0:
            raise IndexError(
                f"Negative index {start} out of bounds for axis of size {size_in_axis}"
            )

        start = size_in_axis + start

    if stop < 0:
        if stop + size_in_axis < 0:
            raise IndexError(
                f"Negative index {stop} out of bounds for axis of size {size_in_axis}"
            )

        stop = size_in_axis + stop

    # Done
    return start, stop - start


def _string_index_to_start_and_extent(
    index: str | int | slice,
    column_names: list[str] | None,
    size_in_axis: int,
    axis: t.Literal["x", "y"],
) -> tuple[
    int | t.Literal["header"],
    int,
]:
    """
    Same as `_index_to_start_and_extent`, but with support for string indices.
    """
    # If passed a string, select the entire column
    if isinstance(index, str):
        if column_names is None:
            raise ValueError(
                "Cannot index into this table using a column name, since it doesn't have any headers"
            )

        try:
            left = column_names.index(index)
        except ValueError:
            raise KeyError(f"This table doesn't have a column named {index!r}")

        return left, 1

    # Otherwise delegate to the other function
    return _index_to_start_and_extent(
        index,
        size_in_axis,
        axis,
    )


def _indices_to_rectangle(
    index: str | tuple[int | slice | str, str | int | slice],
    column_names: list[str] | None,
    data_width: int,
    data_height: int,
) -> tuple[
    int,
    int | t.Literal["header"],
    int,
    int,
]:
    # Get the raw x & y indices
    if isinstance(index, str):
        index_y = slice(None)
        index_x = index
    else:
        if not isinstance(index, tuple):
            raise ValueError(
                f"Table indices should be a tuple of two elements, not {index!r}"
            )
        elif len(index) != 2:
            raise ValueError(
                f"Table indices should have exactly two elements, not {len(index)}"
            )

        index_y, index_x = index

    # Get the index as a tuple (top, left, height, width)
    top, height = _index_to_start_and_extent(
        index_y,
        data_height,
        axis="y",
    )

    left, width = _string_index_to_start_and_extent(
        index_x,
        column_names,
        data_width,
        axis="x",
    )

    assert isinstance(left, int), left  # Left can't be "header"

    return left, top, width, height


def _convert_iterable(
    values: t.Iterable[t.Any],
    date_format_string: str,
) -> list[NormalizedTableValue]:
    """
    Converts an iterable of table values into a list of normalized ones. This is
    used to ensure that all table values are JSON-serializable and supported by
    the backend.
    """
    result: list[NormalizedTableValue] = []

    for value in values:
        # Some values can be kept as-is
        if isinstance(value, (int, float, str)):
            result.append(value)

        # Format dates using the session's date format
        elif isinstance(value, date):
            result.append(value.strftime(date_format_string))

        # Apply `str` to everything else
        else:
            result.append(str(value))

    return result


def _data_to_columnar(
    data: pandas.DataFrame
    | polars.DataFrame
    | numpy.ndarray
    | t.Mapping[str, t.Iterable[t.Any]]
    | t.Iterable[t.Iterable[t.Any]],
    date_format_string: str,
) -> tuple[
    list[str] | None,
    list[list[NormalizedTableValue]],
]:
    """
    Converts table data in any of the supported data formats into a standardized
    one. The result is a list of headers and table columns.
    """

    headers: list[str] | None = None
    columns: list[list[NormalizedTableValue]] = []

    # DataFrame
    #
    # Use narwhals to abstract away the dataframe provider
    if isinstance(data, maybes.DATAFRAME_TYPES):
        nw_data = nw.from_native(data)
        headers = nw_data.columns

        for col_name in nw_data.columns:
            # If the entire column is a supported type, use it as is.
            col = nw_data[col_name]

            if isinstance(col.dtype, narwhals.dtypes.NumericType):
                columns.append(col.to_list())
                continue

            # Otherwise walk all values and convert them one by one
            columns.append(_convert_iterable(col, date_format_string))

    # NumPy array
    #
    # These are neatly organized, just need to get the contents as columns
    elif isinstance(data, maybes.NUMPY_ARRAY_TYPES):
        import numpy as np

        if data.ndim != 2:
            raise ValueError("The table data must be two-dimensional")

        if not np.issubdtype(data.dtype, np.number):
            raise ValueError("The table data must be numeric")

        for ii in range(data.shape[1]):
            columns.append(data[:, ii].tolist())  # type: ignore

    # Mapping
    #
    # The headers are trivially available. The columns can also be used as-is,
    # but care must be taken to make sure they are all the same length.
    elif isinstance(data, t.Mapping):
        headers = list(data.keys())  # type: ignore (wtf?)
        column_lengths = set()

        for raw_column in data.values():
            converted_column = _convert_iterable(raw_column, date_format_string)
            columns.append(converted_column)
            column_lengths.add(len(converted_column))

        if len(column_lengths) > 1:
            raise ValueError("All table columns must have the same length")

    # Iterable of iterables
    #
    # There are no headers. The rows need to be transposed and must all be of
    # equal length.
    else:
        data = t.cast(t.Iterable[t.Iterable[t.Any]], data)
        row_lengths = set()

        for raw_row in data:
            converted_row = _convert_iterable(raw_row, date_format_string)
            columns.append(converted_row)
            row_lengths.add(len(converted_row))

        if len(row_lengths) > 1:
            raise ValueError("All table rows must have the same length")

        # Black magic to transpose the data
        columns = list(map(list, zip(*columns)))

    # Done
    return headers, columns



import io
import re
import shutil
import string
import typing as t
from pathlib import Path

import introspection
import isort
import revel
from revel import fatal, print, success

import rio.cli
import rio.snippets

from .. import project_config

__all__ = [
    "create_project",
]


def class_name_from_snippet(snip: rio.snippets.Snippet) -> str:
    """
    Given a file name, determine the name of the class that is defined in it.

    e.g. `my_component.py` -> `MyComponent`
    """
    assert snip.name.endswith(".py"), snip.name

    parts = snip.name[:-3].split("_")
    return "".join(part.capitalize() for part in parts)


def write_init_file(
    fil: t.IO, snippets: t.Iterable[rio.snippets.Snippet]
) -> None:
    """
    Write an `__init__.py` file that imports all of the snippets.

    e.g. if told to import snippets `foo.py` and `bar.py`, it will write:

    ```
    from .foo import Foo
    from .bar import Bar
    ```
    """
    for snippet in snippets:
        assert snippet.name.endswith(".py"), snippet.name
        module_name = snippet.name[:-3]
        class_name = class_name_from_snippet(snippet)
        fil.write(f"from .{module_name} import {class_name}\n")


def generate_root_init(
    out: t.TextIO,
    *,
    raw_name: str,
    project_type: t.Literal["app", "website"],
    template: rio.snippets.ProjectTemplate,
    as_live_example: bool,
) -> None:
    """
    Generate the `__init__.py` file for the main module of the project.
    """
    assert len(template.page_snippets) > 0, template.page_snippets

    # Imports
    default_theme = rio.Theme.from_colors()
    buffer = io.StringIO()

    buffer.write(
        """
from __future__ import annotations

from pathlib import Path
import typing as t

import rio

from . import components as comps
    """.strip()
    )

    # Additional imports
    try:
        additional_imports = template.root_init_snippet.get_section(
            "additional-imports"
        )
    except KeyError:
        needs_isort = False
    else:
        buffer.write("\n")
        buffer.write(additional_imports)
        buffer.write("\n")
        needs_isort = True

    buffer.write("\n\n")

    # Additional code
    try:
        additional_code = template.root_init_snippet.get_section(
            "additional-code"
        )
    except KeyError:
        pass
    else:
        buffer.write("\n")
        buffer.write(additional_code)
        buffer.write("\n\n")

    # If this is a live example, the template may optionally provide additional
    # code.
    #
    # If this is the case, the code will provide a function called
    # `on_demo_session_start`, which must be used instead of the regular session
    # start. This allows the demo code to clear the session for each user.
    #
    # This supersedes any previous `on_session_start` function - the demo code
    # can call that function itself if it needs to.
    on_session_start_function_name: str | None = template.on_session_start

    if as_live_example:
        try:
            additional_demo_code = template.root_init_snippet.get_section(
                "additional-demo-code"
            )
        except KeyError:
            pass
        else:
            buffer.write("\n")
            buffer.write(additional_demo_code)
            buffer.write("\n\n")

            on_session_start_function_name = "on_demo_session_start"

    # Theme generation if not provided by the template
    if template.theme is None:
        buffer.write(
            f"""
# Define a theme for Rio to use.
#
# You can modify the colors here to adapt the appearance of your app or website.
# The most important parameters are listed, but more are available! You can find
# them all in the docs
#
# https://rio.dev/docs/api/theme
theme = rio.Theme.from_colors(
    primary_color=rio.Color.from_hex("{default_theme.primary_color.hexa}"),
    secondary_color=rio.Color.from_hex("{default_theme.secondary_color.hexa}"),
    mode="light",
)


""".lstrip()
        )

    # App generation
    buffer.write("# Create the Rio app\n")
    buffer.write("app = rio.App(\n")

    # In live demos, use the template name as app name
    if as_live_example:
        buffer.write(f"    name={template.name!r},\n")
    else:
        buffer.write(f"    name={raw_name!r},\n")

    # Optional: `default_attachments`
    if template.default_attachments is not None:
        buffer.write(
            f"    default_attachments=[{', '.join(template.default_attachments)}],\n"
        )

    # Optional: `on_app_start`
    if template.on_app_start is not None:
        buffer.write(
            f"    # This function will be called once the app is ready.\n"
            f"    #\n"
            f"    # `rio run` will also call it again each time the app is reloaded.\n"
            f"    on_app_start={template.on_app_start},\n"
        )

    # Optional: `on_session_start`
    if on_session_start_function_name is not None:
        if project_type == "app":
            buffer.write(
                "    # This function will be called when a user opens the app. Its\n"
                "    # mostly used for web apps, where each user starts a new\n"
                "    # session when they connect.\n"
            )
        else:
            buffer.write(
                "    # This function will be called each time a user connects\n"
            )

        buffer.write(
            f"    on_session_start={on_session_start_function_name},\n"
        )

    # Optional: `root_component`
    if template.root_component is not None:
        buffer.write(
            f"    # You can optionally provide a root component for the app. By default,\n"
            f"    # Rio's default navigation is used. By providing your own component, you\n"
            f"    # can create components which stay put while the user navigates between\n"
            f"    # pages, such as a navigation bar or footer.\n"
            f"    #\n"
            f"    # When you do this, make sure your component contains a `rio.PageView`\n"
            f"    # so the currently active page is still visible.\n"
            f"    build=comps.{template.root_component},\n"
        )

    # Pass in the theme
    if template.theme is not None:
        buffer.write(
            f"    # You can also provide a custom theme for the app. This theme will\n"
            f"    # override Rio's default.\n"
            f"    theme={template.theme},\n"
        )
    elif as_live_example:
        # In live examples, enable both light and dark mode
        buffer.write("    theme=rio.Theme.pair_from_colors(),\n")
    else:
        buffer.write("    theme=theme,\n")

    # Configure the assets directory
    buffer.write('    assets_dir=Path(__file__).parent / "assets",\n')
    buffer.write(")\n\n")

    # Due to imports coming from different sources they're often not sorted. Fix
    # that now.
    if needs_isort:
        formatted_code = isort.code(buffer.getvalue())
        out.write(formatted_code)
    else:
        out.write(buffer.getvalue())


def strip_invalid_filename_characters(name: str) -> str:
    """
    Given a name, strip any characters that are not allowed in a filename.
    """
    return re.sub(r'[<>:"/\\|?*]', "", name)


def derive_module_name(raw_name: str) -> str:
    """
    Given an arbitrary string, derive similar and valid all_lower Python module
    identifier from it.
    """
    # Convert to lower_case
    name = introspection.convert_case(raw_name, "snake")

    # Strip any invalid characters
    name = "".join(c for c in name if c.isidentifier() or c in string.digits)

    # Since modules are written to files, the name also has to be a valid file
    # name
    name = strip_invalid_filename_characters(name)

    # Identifiers cannot start with a digit
    while name and name[0].isdigit():
        name = name[1:]

    # This could've resulted in an empty string
    if not name:
        name = "rio_app"

    # Done
    return name


def generate_readme(
    out: t.TextIO,
    raw_name: str,
    template: rio.snippets.ProjectTemplate,
) -> None:
    out.write(
        f"""# {raw_name}

This is a placeholder README for your project. Use it to describe what your
project is about, to give new users a quick overview of what they can expect.

_{raw_name.capitalize()}_ was created using [Rio](https://rio.dev/), an easy to
use app & website framework for Python.
"""
    )

    # Include the template's README
    if template.name != "Empty":
        out.write(
            f"""
This project is based on the `{template.name}` template.

## {template.name}

{template.description_markdown_source}
"""
        )


def write_component_file(
    out: t.TextIO,
    snip: rio.snippets.Snippet,
    import_depth: int,
) -> None:
    """
    Writes the Python file containing a component or page to the given file.
    """
    # Common imports
    buffer = io.StringIO()

    dots = "." * (import_depth + 1)

    buffer.write(
        f"""from __future__ import annotations

from dataclasses import KW_ONLY, field
import typing as t

import rio

from {dots} import components as comps

"""
    )

    # Additional, user-defined imports
    try:
        additional_imports = snip.get_section("additional-imports")
    except KeyError:
        needs_isort = False
    else:
        buffer.write(additional_imports)
        buffer.write("\n")
        needs_isort = True

    # The component proper
    buffer.write(snip.get_section("component"))

    # Due to imports coming from different sources they're often not sorted.
    # -> Apply `isort`
    if needs_isort:
        formatted_code = isort.code(buffer.getvalue())
        out.write(formatted_code)
    else:
        out.write(buffer.getvalue())


def generate_dependencies_file(
    project_dir: Path, dependencies: dict[str, str]
) -> None:
    """
    Writes a `requirements.txt` file with the given dependencies. Does nothing
    if there are no dependencies.
    """
    # Anything to do?
    if not dependencies:
        return

    # requirements.txt
    with open(project_dir / "requirements.txt", "w", encoding="utf-8") as out:
        for package, version_specifier in dependencies.items():
            out.write(f"{package}{version_specifier}\n")


def create_project(
    *,
    raw_name: str,
    type: t.Literal["app", "website"],
    template_name: rio.snippets.AvailableTemplatesLiteral,
    target_parent_directory: Path,
    as_live_example: bool = False,
) -> None:
    """
    Create a new project with the given name. This will directly interact with
    the terminal, asking for input and printing output.

    ## Parameters

    `as_live_example`: This makes some changes to the created project to make it
        more suitable for live examples. For example, additional code can be run
        on session start to clean up the app for every user.
    """

    # Derive a valid module name
    module_name = derive_module_name(raw_name)

    # The project directory is called the same, but in kebab-case
    dashed_name = module_name.replace("_", "-")

    # Find the template
    for template in rio.snippets.get_project_templates(include_empty=True):
        if template.name == template_name:
            break
    else:
        assert False, (
            f"Received invalid template name `{template_name}`. This shouldn't be possible if the types are correct."
        )

    # Create the target directory
    project_dir = target_parent_directory / dashed_name
    project_dir.mkdir(parents=True, exist_ok=True)

    # If the project directory already exists it must be empty
    if any(project_dir.iterdir()):
        fatal(
            f"The project directory `{project_dir}` already exists and is not empty"
        )

    # Generate /rio.toml
    project_config.RioProjectConfig.create(
        project_dir / "rio.toml",
        main_module=module_name,
        project_type=type,
    )

    # Create the main module and its subdirectories
    main_module_dir = project_dir / module_name
    assets_dir = main_module_dir / "assets"
    components_dir = main_module_dir / "components"
    pages_dir = main_module_dir / "pages"

    main_module_dir.mkdir(parents=True, exist_ok=True)
    assets_dir.mkdir()
    components_dir.mkdir()
    pages_dir.mkdir()

    # Generate /assets/*
    for snip in template.asset_snippets:
        source_path = snip.file_path
        target_path = assets_dir / source_path.name
        shutil.copyfile(source_path, target_path)

    # Generate /components/*.py
    for snip in template.component_snippets:
        target_path = components_dir / snip.name

        with target_path.open("w", encoding="utf-8") as f:
            write_component_file(
                f,
                snip,
                import_depth=1,
            )

    # Generate pages/*.py
    #
    # Pages are more complicated, because they may be placed in subdirectories.
    # Get this page's relative path.
    template_pages_dir = template.root_init_snippet.file_path.parent / "pages"

    for snip in template.page_snippets:
        rel_path = snip.file_path.relative_to(template_pages_dir)
        target_path = pages_dir / rel_path

        target_path.parent.mkdir(parents=True, exist_ok=True)

        with target_path.open("w", encoding="utf-8") as f:
            write_component_file(
                f,
                snip,
                import_depth=len(rel_path.parts),
            )

    # Generate /*.py
    for snip in template.other_python_files:
        source_string = snip.stripped_code()
        target_path = main_module_dir / snip.name

        with target_path.open("w", encoding="utf-8") as f:
            f.write(source_string)

    # Generate /project/__init__.py
    with open(main_module_dir / "__init__.py", "w", encoding="utf-8") as fil:
        generate_root_init(
            out=fil,
            raw_name=raw_name,
            project_type=type,
            template=template,
            as_live_example=as_live_example,
        )

    # Generate /project/components/__init__.py
    with open(
        main_module_dir / "components" / "__init__.py", "w", encoding="utf-8"
    ) as f:
        write_init_file(f, template.component_snippets)

    # Generate a file specifying all dependencies, if there are any
    generate_dependencies_file(project_dir, template.dependencies)

    # Generate README.md
    with open(project_dir / "README.md", "w", encoding="utf-8") as f:
        generate_readme(f, raw_name, template)

    # Applications require a `__main__.py` as well
    if type == "app":
        with open(main_module_dir / "__main__.py", "w", encoding="utf-8") as f:
            f.write(
                f"""
# Make sure the project is in the Python path
import sys
from pathlib import Path
sys.path.insert(0, str(Path(__file__).absolute().parent.parent))


# Import the main module
import {module_name}

# Run the app
{module_name}.app.run_in_window()
"""
            )

    # Report success
    #
    # TODO: Other tools like poetry!? Add a command to activate the venv?
    print()
    success("The project has been created!")
    success(f"You can find it at `{project_dir.resolve()}`")
    print()
    print("To see your new project in action, run the following commands:")
    print()
    print(f"[dim]>[/] cd {revel.shell_escape(project_dir.resolve())}")

    if template.dependencies:
        print(
            "[dim]>[/] python -m pip install -r requirements.txt  [bold]# Don't forget to install dependencies![/]"
        )

    print("[dim]>[/] rio run")
